"""
Luna accounts client
"""

from numbers import Real
from typing import Awaitable, List, Optional, Union

import aiohttp

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..common.base_api import BaseHTTPAPI
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from .http_objs import TokenPermissions


class AccountsApi(BaseHTTPAPI):
    """
    Class for request to luna-accounts.
    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5170",
        api: Optional[int] = 1,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def createAccount(
        self,
        login: str,
        password: str,
        accountType: Literal["admin", "advanced_user", "user"],
        description: Optional[str] = None,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new account in luna-accounts.

        Args:
            login: account login
            password: account password
            accountType: account type
            description: account description
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created account id will be returned.

        """
        data = dict(login=login, password=password, account_type=accountType)
        if description is not None:
            data["description"] = description
        payload = RequestPayload.buildJson(body=data)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/accounts", "POST", body=payload, **kwargs)

    def getAccount(
        self,
        accountId: str,
        targets: Optional[List[str]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get account from luna-accounts

        Args:
            accountId: account id
            targets: targets to get

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with account will be returned.

        """
        return self.makeRequest(
            f"{self.baseUri}/accounts/{accountId}", "GET", queryParams={"targets": targets}, **kwargs
        )

    def getAccounts(
        self,
        page: int = 1,
        pageSize: int = 10,
        login: Optional[str] = None,
        accountType: Optional[Literal["admin", "advanced_user", "user"]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        targets: Optional[List[str]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get accounts from luna-accounts

        Args:
            page: page, default 1
            pageSize: page size, default 10
            login: account login
            accountType: account type
            createTimeLt: upper bound for account create_time
            createTimeGte: lower bound for account create_time
            targets: targets to get

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with accounts will be returned.

        """
        queries = {
            "page": page,
            "page_size": pageSize,
            "login": login,
            "account_type": accountType,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
            "targets": targets,
        }
        return self.makeRequest(f"{self.baseUri}/accounts", "GET", queryParams=queries, **kwargs)

    def patchAccount(
        self,
        accountId: str,
        login: Optional[str] = None,
        password: Optional[str] = None,
        accountType: Optional[Literal["admin", "advanced_user", "user"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch account in luna-accounts.

        Args:
            accountId: account id
            login: account login
            password: account password
            accountType: account type
            description: account description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` 204 status code will be returned.
        """
        data = {}
        if login is not None:
            data["login"] = login
        if password is not None:
            data["password"] = password
        if accountType is not None:
            data["account_type"] = accountType
        if description is not None:
            data["description"] = description
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/accounts/{accountId}", "PATCH", body=payload, **kwargs)

    def deleteAccount(self, accountId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete account in luna-accounts.

        Args:
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` 204 status code will be returned.
        """
        return self.makeRequest(f"{self.baseUri}/accounts/{accountId}", "DELETE", **kwargs)

    def createToken(
        self,
        accountId: str,
        grantAllPermissions: int = 0,
        tokenPermissions: TokenPermissions | None = None,
        expirationTime: Optional[str] = None,
        visibilityArea: Optional[Literal["all", "account"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new token in luna-accounts.

        Args:
            accountId: account id
            grantAllPermissions: grant all permissions to token
            tokenPermissions: token permissions
            expirationTime: token expiration time
            visibilityArea: token visibility area
            description: account description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created token id will be returned.

        """
        data = {"expiration_time": expirationTime}
        if tokenPermissions is not None:
            data["permissions"] = tokenPermissions.asDict()
        if description is not None:
            data["description"] = description
        if visibilityArea is not None:
            data["visibility_area"] = visibilityArea
        payload = RequestPayload.buildJson(body=data)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(
            f"{self.baseUri}/tokens",
            "POST",
            body=payload,
            queryParams={"grant_all_permissions": grantAllPermissions},
            **kwargs,
        )

    def getTokens(
        self,
        page: int = 1,
        pageSize: int = 10,
        accountId: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        permissionsTargets: Optional[Literal["all", "custom", "standard"]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tokens from luna-accounts

        Args:
            accountId: account id
            page: page, default 1
            pageSize: page size, default 10
            createTimeLt: upper bound for token create_time
            createTimeGte: lower bound for token create_time
            permissionsTargets: permissions targets to return

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with tokens will be returned.

        """
        queries = {
            "account_id": accountId,
            "page": page,
            "page_size": pageSize,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
            "permissions_targets": permissionsTargets,
        }
        return self.makeRequest(f"{self.baseUri}/tokens", "GET", queryParams=queries, **kwargs)

    def replaceToken(
        self,
        tokenId: str,
        grantAllPermissions: int = 0,
        tokenPermissions: TokenPermissions | None = None,
        accountId: Optional[str] = None,
        expirationTime: Optional[str] = None,
        visibilityArea: Optional[Literal["all", "account"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Replace token in luna-accounts

        Args:
            tokenId: token id
            grantAllPermissions: grant all permissions to token
            tokenPermissions: token permissions
            accountId: account id
            expirationTime: token expiration time
            description: account description
            visibilityArea: token visibility area

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created token id will be returned.

        """
        data = {"expiration_time": expirationTime}
        if tokenPermissions is not None:
            data["permissions"] = tokenPermissions.asDict()
        if description is not None:
            data["description"] = description
        if visibilityArea is not None:
            data["visibility_area"] = visibilityArea
        payload = RequestPayload.buildJson(body=data)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(
            f"{self.baseUri}/tokens/{tokenId}",
            "PUT",
            body=payload,
            queryParams={"grant_all_permissions": grantAllPermissions},
            **kwargs,
        )

    def getToken(
        self,
        accountId: str,
        tokenId: str,
        permissionsTargets: Optional[Literal["all", "custom", "standard"]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get token from luna-accounts

        Args:
            accountId: account id
            tokenId: token id
            permissionsTargets: permissions targets to return

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with token will be returned.

        """
        queries = {"account_id": accountId, "permissions_targets": permissionsTargets}
        return self.makeRequest(f"{self.baseUri}/tokens/{tokenId}", "GET", queryParams=queries, **kwargs)

    def deleteToken(
        self, tokenId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete token from luna-accounts

        Args:
            tokenId: token id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return nothing in text field

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tokens/{tokenId}", "DELETE", queryParams=queries, **kwargs)

    def verifyAccount(
        self,
        accountId: Optional[str] = None,
        token: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        permissionsTargets: Optional[Literal["all", "custom", "standard"]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Verify account data in luna-accounts.

        Args:
            accountId: account id
            token: account JWT token
            login: account login
            password: account password
            permissionsTargets: targets for permissions in response
        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` 204 status code will be returned.
        """
        queries = {"permissions_targets": permissionsTargets}
        data = {}
        if login is not None:
            data["login"] = login
        if password is not None:
            data["password"] = password
        if accountId is not None:
            data["account_id"] = accountId
        if token is not None:
            data["token"] = token
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/verifier", "POST", body=payload, queryParams=queries, **kwargs)

    def countAccounts(self, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Count accounts in luna-accounts.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with accounts count will be returned.
        """
        return self.makeRequest(f"{self.baseUri}/accounts/count", "GET", **kwargs)
