"""
Module realize faces mixin
"""

from typing import Awaitable, List, Optional, Union

from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.lunavl.mixins.base import MixinStub


class FacesMixin(MixinStub):
    """Faces Mixin to work with faces."""

    def createFace(
        self,
        externalId: Optional[str] = None,
        attributeId: Optional[str] = None,
        userData: Optional[str] = None,
        avatar: Optional[str] = None,
        lists: Optional[List[str]] = None,
        eventId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new face

        Args:
            externalId: user-defined external id for face
            attributeId: an attribute id associated with the face
            userData: user data
            avatar: avatar for face, references to a sample or a user-defined reference
            lists: luna lists for linking face with them
            eventId: event id associated with creating the face

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: info about created face

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> response = client.createFace(
        >>>     externalId="2xQ2gprbMePw1s9gw9fvA==", userData="test_data05",
        >>>     attributeId="af327028-e3b2-4622-b21f-ba5f8180da06",
        >>>     avatar="https://upload.wikimedia.org/wikipedia/commons/thumb/c/c4/Abbey_Lee_2015.jpg/"
        >>>         "800px-Abbey_Lee_2015.jpg",
        >>>     lists=["cd59b16e-bdd7-49a4-b90a-b62c804580b4"])
        >>> print(response.statusCode)
        201
        >>> print(response.json)
        {
           'face_id': '08c26dc6-feff-4382-b1a7-8b9c25134e32',
           'url': '/1/faces/08c26dc6-feff-4382-b1a7-8b9c25134e32'
        }

        """
        bodyJson = {
            "external_id": externalId,
            "attribute_id": attributeId,
            "user_data": userData,
            "avatar": avatar,
            "lists": lists,
            "event_id": eventId,
        }
        payload = RequestPayload.buildJson(body={k: v for k, v in bodyJson.items() if v is not None})
        return self.makeRequest(f"{self.baseUri}/faces", "POST", body=payload, **kwargs)

    def getFaces(
        self,
        targets: Optional[List[str]] = None,
        page: int = 1,
        pageSize: int = 10,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        externalId: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Getting faces which satisfy filters

        Args:
            targets: list of face fields to receive instead of full faces
            page: The number of page
            pageSize: The numbers of items on page
            timeLt: Upper bound for object create_time
            timeGte: Lower bound for object create_time
            faceIds: list of coma separated faces ids
            externalId: external id
            userData: Find all objects with user_data which "like" of this parameter
            listId: id of list that contains faces

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with faces info

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> response = client.getFaces(timeGte="2018-08-10", userData="fox1991")
        >>> print(response.json)
        [
         {
           "face_id": "426542d6-5509-4e5b-8a01-e2abd5c0a8c6",
           "external_id": "2xQ2gprbMUePw1s9gw9fvA==",
           "attribute_id": "b668c4a5-2191-476e-a261-3b4f9ce2e25e",
           "user_data": "fox1991",
           "create_time": "2018-08-11T09:11:41.674Z",
           "lists": [
             "429b0e78-4616-426a-b57f-02baa72d638d",
             "6d037c33-31ec-4d73-b3b3-ec80b09446c2"
           ],
           "avatar": "https://st.kp.yandex.net/images/actor_iphone/iphone360_460188.jpg",
           "event_id": "f9687459-986b-406d-9c1f-0d6289be5256"
         }
        ]
        >>> print(response.statusCode)
        200

        """
        queries = {
            "targets": targets,
            "page": page,
            "page_size": pageSize,
            "time__lt": timeLt,
            "time__gte": timeGte,
            "face_ids": faceIds,
            "external_id": externalId,
            "user_data": userData,
            "list_id": listId,
        }
        return self.makeRequest(f"{self.baseUri}/faces", "GET", queryParams=queries, **kwargs)

    def countFaces(
        self,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        externalId: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Getting count of faces which satisfy filters

        Args:
            timeLt: Upper bound for object create_time
            timeGte: Lower bound for object create_time
            faceIds: list of coma separated faces ids
            externalId: external id
            userData: Find all objects with user_data which "like" of this parameter
            listId: id of list that contains faces

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: number

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> response = client.countFaces(timeGte="2018-08-10", userData="fox1991")
        >>> print(response.json)
        5
        >>> print(response.statusCode)
        200

        """
        queries = {
            "time__lt": timeLt,
            "time__gte": timeGte,
            "face_ids": faceIds,
            "external_id": externalId,
            "user_data": userData,
            "list_id": listId,
        }
        return self.makeRequest(f"{self.baseUri}/faces/count", "GET", queryParams=queries, **kwargs)

    def getFace(
        self, faceId: str, targets: Optional[List[str]] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face by id

        Args:
            targets: list of face fields to receive instead of full face
            faceId: string <uuid4> (uuid4) ^[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$/

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with face information

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> response = client.getFace('51fcd44c-63f8-47ee-894d-31fbbbe1e8b7')
        >>> print(response.statusCode)
        200
        >>> print(response.json)
        {
          "face_id": "51fcd44c-63f8-47ee-894d-31fbbbe1e8b7",
          "account_id": "00000000-0000-4000-b000-000000000000",
          "user_data": "some_string",
          "create_time": "2018-09-19T20:23:58.932265+03:00",
          "event_id": None,
          "external_id": "2xQ2gprbMUePw1s9gw9fvA==",
          "avatar": "https://upload.wikimedia.org/wikipedia/commons/thumb/c/c4/Abbey_Lee_2015.jpg/
            800px-Abbey_Lee_2015.jpg",
          "lists": [],
          "attribute_id": None
        }

        """
        queries = {"targets": targets}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "GET", queryParams=queries, **kwargs)

    def patchFace(
        self,
        faceId: str,
        externalId: Optional[str] = None,
        attributeId: Optional[str] = None,
        userData: Optional[str] = None,
        avatar: Optional[str] = None,
        eventId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update fields of face (user_data, external_id, attribute_id)

        Args:
            faceId: face id; uuid4;
            externalId: user-defined external id for face
            attributeId: an attribute id associated with the face
            userData: user data
            avatar: avatar for face, references to a sample or a user-defined reference
            eventId: event id associated with creating the face

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: nothing

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> response = client.patchFace(faceId='08c26dc6-feff-4382-b1a7-8b9c25134e32',
        >>>                             userData="test_data05_patched")
        >>> print(response.statusCode)
        204
        >>> print(response.text)

        """
        bodyJson = {
            "external_id": externalId,
            "attribute_id": attributeId,
            "user_data": userData,
            "avatar": avatar,
            "event_id": eventId,
        }
        payload = RequestPayload.buildJson(body={k: v for k, v in bodyJson.items() if v is not None})
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "PATCH", body=payload, **kwargs)

    def deleteFace(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete face by id

        Args:
            faceId: string <uuid4> (uuid4) ^[a-f0-9]{8}-[a-f0-9]{4}-4[a-f0-9]{3}-[89ab][a-f0-9]{3}-[a-f0-9]{12}$/

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: nothing

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> response = client.deleteFace(faceId='51fcd44c-63f8-47ee-894d-31fbbbe1e8b7')
        >>> print(response.text)

        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "DELETE", **kwargs)

    def checkFace(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check a face with id=face_id existence

        Args:
            faceId: face id (uuid4)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: nothing

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> response = client.checkFace('51fcd44c-63f8-47ee-894d-31fbbbe1e8b7')
        >>> print(response.text)

        >>> print(response.statusCode)
        200

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "HEAD", **kwargs)

    def getFaceAttributes(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face attributes

        Args:
            faceId: face id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with face information

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient()
        >>> response = client.getFaceAttributes('9ed471d4-0130-4fd5-8053-273f7fd15ff8')
        >>> print(response.statusCode)
        200
        >>> print(response.json)
        {"attribute_id":"bf1deca9-261c-490a-8608-1b6178aab361","gender":0,"age":27,"ethnicity":4}

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes", "GET", **kwargs)

    def getFaceSamples(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get samples which are associated with the face attributes (gender, age, ethnicity, descriptor).

        Args:
            faceId: face Id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse`: json with basic attributes

        >>> from luna3.backport4.httpclient import LunaHttpClient
        >>>
        >>> client = LunaHttpClient(accountId='02bbe8fe-7dd2-46bf-85c9-494693a9d30b')
        >>> response = client.getFaceSamples("7d60014f-d7c6-4398-a629-dc2616018d47")
        >>> print(response.json)
        [
            {
                "sample_id": "3ba4b5bc-b3cd-4a06-b6bf-254db69d63ed",
                "url": "/1/samples/3ba4b5bc-b3cd-4a06-b6bf-254db69d63ed"
            },
            {
                "sample_id": "946f92aa-90cf-4a64-aa58-b896ee1ab991",
                "url": "/1/samples/946f92aa-90cf-4a64-aa58-b896ee1ab991"
            }
        ]
        >>> print(response.statusCode)
        200

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes/samples", "GET", **kwargs)
