"""
Helper functions for services
"""

import io
from typing import Iterable, List, Optional, Union

import aiohttp

from ..common.http_objs import BinaryDescriptorData, BinaryImage, Policies, SampleForDetection, UrlForDetection
from ..common.requests import RequestPayload


def makeContentType(
    inputData: Union[BinaryImage, BinaryDescriptorData, List[Union[BinaryImage, BinaryDescriptorData]]],
    multipartBoundary: str,
) -> dict:
    """
    Returns dict with suitable Content-Type header

    Args:
        inputData: raw input data (images and/or raw descriptors)
        multipartBoundary (str): multipart boundary

    Returns:
        dict for headers

    """
    headers = {}
    if isinstance(inputData, list):
        headers["Content-Type"] = "multipart/form-data; boundary={}".format(multipartBoundary)
    else:
        headers["Content-Type"] = inputData.mimetype
    return headers


def checkContentType(
    inputData: Union[BinaryImage, BinaryDescriptorData, List[Union[BinaryImage, BinaryDescriptorData]]],
    allowedTypes: Iterable[str],
) -> None:
    """ "
    The function check if mimetype of images is in allowedTypes

    Args:
        inputData: raw input data (images and/or raw descriptors)
        allowedTypes (tuple): tuple of strings with allowed mimetypes

    Raises:
        ValueError: if body has unsupported mimetype
    """

    if isinstance(inputData, list):
        for image in inputData:
            if image.mimetype not in allowedTypes:
                raise ValueError("wrong mimetype, available mimetypes are: {}".format(allowedTypes))
    else:
        if inputData.mimetype not in allowedTypes:
            raise ValueError("wrong mimetype, available mimetypes are: {}".format(allowedTypes))


def getMultipartBody(
    images: Iterable[Union[BinaryImage, BinaryDescriptorData]],
    multipartBoundary: str,
    policies: Optional[Union[Policies, dict]] = None,
    mutualMeta: Optional[dict] = None,
) -> Union[aiohttp.MultipartWriter, bytes]:
    """
    Get body for multipart request
    Args:
        images: list of images | optionally with bounding box(es) or raw descriptors
        policies: optionally policies for dynamic handler
        mutualMeta: user-defined event mutual metadata
        multipartBoundary: multipart boundary

    Returns:
        multipart request body
    """
    faceBoundingBoxes = []
    bodyBoundingBoxes = []
    imagesDetectTimes = []
    imagesDetectTSs = []
    imagesOrigins = []
    imagesMeta = []

    with aiohttp.MultipartWriter("form-data", multipartBoundary) as mpWriter:
        for img in images:
            imagePart = mpWriter.append(io.BytesIO(img.body), img.httpMultipartHeaders)
            imagePart.set_content_disposition("form-data", name="image", filename=img.filename, quote_fields=False)
            if isinstance(img, BinaryDescriptorData):
                continue
            if img.faceBoundingBoxList is not None:
                faceBoundingBoxes.append(
                    {
                        "filename": img.filename,
                        "face_bounding_boxes": [boundingBox.asDict() for boundingBox in img.faceBoundingBoxList],
                    }
                )
            if img.bodyBoundingBoxList is not None:
                bodyBoundingBoxes.append(
                    {
                        "filename": img.filename,
                        "body_bounding_boxes": [boundingBox.asDict() for boundingBox in img.bodyBoundingBoxList],
                    }
                )
            if img.detectTime is not None:
                imagesDetectTimes.append({"filename": img.filename, "detect_time": img.detectTime})
            if img.imageOrigin is not None:
                imageOriginData = {"filename": img.filename, "image_origin": img.imageOrigin}
                if img.imageOriginMimetype is not None:
                    imageOriginData["mimetype"] = img.imageOriginMimetype
                imagesOrigins.append(imageOriginData)
            if img.detectTs is not None:
                imagesDetectTSs.append({"filename": img.filename, "detect_ts": img.detectTs})
            if img.meta is not None:
                imagesMeta.append({"filename": img.filename, "meta": img.meta})

        if faceBoundingBoxes:
            part = mpWriter.append_json(faceBoundingBoxes)
            part.set_content_disposition("form-data", name="face_bounding_boxes", quote_fields=False)

        if bodyBoundingBoxes:
            part = mpWriter.append_json(bodyBoundingBoxes)
            part.set_content_disposition("form-data", name="body_bounding_boxes", quote_fields=False)

        if imagesDetectTimes:
            part = mpWriter.append_json(imagesDetectTimes)
            part.set_content_disposition("form-data", name="image_detect_time", quote_fields=False)

        if imagesDetectTSs:
            part = mpWriter.append_json(imagesDetectTSs)
            part.set_content_disposition("form-data", name="image_detect_ts", quote_fields=False)

        if imagesOrigins:
            jsonPart = []
            for imageOrigin in imagesOrigins:
                if isinstance(imageOrigin["image_origin"], str):
                    jsonPart.append(imageOrigin)
                else:
                    originPart = mpWriter.append(
                        io.BytesIO(imageOrigin["image_origin"].body),
                        imageOrigin["image_origin"].httpMultipartHeaders,
                    )
                    originPart.set_content_disposition(
                        "form-data",
                        name="image_origin",
                        filename=imageOrigin["image_origin"].filename,
                        quote_fields=False,
                    )
            if len(jsonPart) > 0:
                part = mpWriter.append_json(jsonPart)
                part.set_content_disposition("form-data", name="image_origin", quote_fields=False)

        if policies is not None:
            part = mpWriter.append_json(policies if isinstance(policies, dict) else policies.asDict())
            part.set_content_disposition("form-data", name="policies", quote_fields=False)

        if mutualMeta and imagesMeta:
            raise RuntimeError("Only one of mutual meta or particular image meta could be passed in multipart.")
        if mutualMeta is not None:
            part = mpWriter.append_json(mutualMeta)
            part.set_content_disposition("form-data", name="meta", quote_fields=False)
        elif imagesMeta:
            part = mpWriter.append_json(imagesMeta)
            part.set_content_disposition("form-data", name="meta", quote_fields=False)

    return mpWriter


def createPayloadWithImages(
    inputData: Union[
        BinaryImage,
        BinaryDescriptorData,
        List[Union[BinaryImage, BinaryDescriptorData]],
        List[UrlForDetection],
        List[str],
        List[SampleForDetection],
    ],
    headers: dict,
    multipartBoundary: str,
    policies: Optional[Union[Policies, dict]] = None,
    meta: Optional[dict] = None,
) -> RequestPayload:
    """
    Prepare data for request with detection or extraction (to send request to detection or handler or sdk)
    Args:
        inputData: input request image data
        headers: request headers
        policies: object with policies information
        meta: user-defined event mutual metadata
        multipartBoundary: multipart boundary

    Returns:
        Tuple with json and body for request and prepared request headers
    Raises:
        ValueError("Cannot find images, or samples, or urls in input args") if cannot find required data
    """
    if isinstance(inputData, BinaryImage):
        if policies or meta is not None:
            payload = RequestPayload.buildMultipart(
                body=getMultipartBody([inputData], multipartBoundary, policies, meta), headers=headers
            )

        elif any(
            item is not None
            for item in (
                inputData.faceBoundingBoxList,
                inputData.bodyBoundingBoxList,
                inputData.detectTime,
                inputData.detectTs,
                inputData.imageOrigin,
                inputData.meta,
            )
        ):
            payload = RequestPayload.buildJson(body=inputData.asJson(), headers=headers)
        else:
            allowedTypes = (
                "image/jpeg",
                "image/png",
                "image/bmp",
                "image/x-portable-pixmap",
                "image/tiff",
                "multipart/form-data",
                "application/json",
                "image/x-jpeg-base64",
                "image/x-png-base64",
                "image/x-bmp-base64",
                "image/x-tiff-base64",
                "image/x-portable-pixmap-base64",
            )
            checkContentType(inputData, allowedTypes)
            payload = RequestPayload(
                body=io.BytesIO(inputData.body), headers={**headers, **makeContentType(inputData, multipartBoundary)}
            )

    elif isinstance(inputData, BinaryDescriptorData):
        if policies is not None:
            payload = RequestPayload.buildMultipart(
                body=getMultipartBody([inputData], multipartBoundary, policies), headers=headers
            )
        else:
            allowedTypes = (
                "application/x-sdk-descriptor",
                "application/x-sdk-descriptor-base64",
                "application/x-vl-xpk",
                "application/x-vl-xpk-base64",
            )
            checkContentType(inputData, allowedTypes)
            payload = RequestPayload(
                body=io.BytesIO(inputData.body), headers={**headers, **makeContentType(inputData, multipartBoundary)}
            )

    elif isinstance(inputData[0], (BinaryImage, BinaryDescriptorData)):
        payload = RequestPayload.buildMultipart(
            body=getMultipartBody(inputData, multipartBoundary, policies, meta), headers=headers
        )

    elif isinstance(inputData[0], UrlForDetection):
        data = {"urls": [urlForDetection.asDict() for urlForDetection in inputData]}
        data.update({"meta": meta} if meta is not None else {})
        payload = RequestPayload.buildJson(body=data, headers=headers)

    elif isinstance(inputData[0], SampleForDetection):
        data = {"samples": [sampleForDetection.asDict() for sampleForDetection in inputData]}
        data.update({"meta": meta} if meta is not None else {})
        payload = RequestPayload.buildJson(body=data, headers=headers)

    elif isinstance(inputData[0], str):
        payload = RequestPayload.buildJson(body={"samples": inputData}, headers=headers)

    else:
        raise ValueError("Cannot find images, or samples, or urls in input args")

    return payload
