""" Raw event object structures """

from dataclasses import dataclass
from typing import Literal, TypeVar

from ..events.http_objs import (
    BaseClass,
    BodyAccessories,
    BodyAggregateEstimations,
    BodyBasicAttributes,
    BodyExtraAttributes,
    BodyUpperAttributes,
    Deepfake,
    Emotions,
    EventMatch,
    Liveness,
    Location,
    MaskEstimations,
    Rect,
)

TP = TypeVar("TP", bound=int)

Point = list[TP]
Landmarks = list[Point]


@dataclass
class FaceOcclusionEstimations(BaseClass):
    """
    Face occlusion estimations container.
    """

    clear: float
    chin: float
    correct: float
    mouth: float
    partially: float
    full: float


@dataclass
class FaceOcclusion(BaseClass):
    """
    Face occlusion container.
    """

    predominantOcclusion: str
    estimations: FaceOcclusionEstimations


@dataclass
class Mask(BaseClass):
    """
    Mask class

    Attributes:
        estimations: mask estimations
        predominantMask: predominant mask name
    """

    estimations: MaskEstimations
    predominantMask: str
    faceOcclusion: FaceOcclusion | None = None


@dataclass
class Gaze(BaseClass):
    """
    Gaze class

    Attributes:
        pitch: angle pitch value
        yaw: angle yaw value
    """

    pitch: float
    yaw: float


@dataclass
class HeadPose(BaseClass):
    """
    HeadPose class

    Attributes:
        pitch: angle pitch value
        yaw: angle yaw value
        roll: angle roll value
    """

    pitch: float
    yaw: float
    roll: float


@dataclass
class MouthAttributes(BaseClass):
    """
    Mouth attributes class

    Attributes:
        opened: mouth opened value
        smile: mouth smile value
        occluded: mouth occluded value
        score: mouth score value
    """

    opened: float
    smile: float
    occluded: float
    score: float


@dataclass
class EyePoint(BaseClass):
    """
    Eye point class

    Attributes:
        state: eye state ["open", "closed", "occluded"]
        irisLandmarks: iris landmarks
    """

    state: str
    irisLandmarks: Landmarks


@dataclass
class EyesAttributes(BaseClass):
    """
    Eyes attributes class

    Attributes:
        leftEye: left eye estimations
        rightEye: right eye estimations
    """

    leftEye: EyePoint
    rightEye: EyePoint


@dataclass
class QualityEstimations(BaseClass):
    """
    Quality estimations class

    Attributes:
        blurriness: blurriness estimation value
        dark: dark estimation value
        illumination: illumination estimation value
        specularity: specularity estimation value
        light: light estimation value
    """

    blurriness: float
    dark: float
    illumination: float
    specularity: float
    light: float


@dataclass
class EthnicityEstimations(BaseClass):
    """
    Ethnicity estimations class

    Attributes:
        asian: asian estimation value
        indian: indian estimation value
        caucasian: caucasian estimation value
        africanAmerican: african-american estimation value
    """

    africanAmerican: float
    asian: float
    indian: float
    caucasian: float


@dataclass
class RawEthnicities(BaseClass):
    """
    Raw ethnicities class

    Attributes:
        predominantEthnicity: predominant ethnicity name
        estimations: ethnicity estimations
    """

    predominantEthnicity: str
    estimations: EthnicityEstimations


@dataclass
class RawBasicAttributes(BaseClass):
    """
    Raw basic attributes class

    Attributes:
        age: age
        ethnicities: ethnicities
        gender: gender (0 - female, 1 - male)
    """

    age: int
    gender: int
    ethnicities: RawEthnicities


@dataclass
class RawFaceSampleAttributes(BaseClass):
    """
    Raw face sample class

    Attributes:
        emotions: face emotions
        mask: face mask
        gaze: face gaze
        headPose: face head pose
        mouthAttributes: face mouth attributes
        eyesAttributes: face eyes attributes
        liveness: face liveness
        deepfake: face deepfake
    """

    emotions: Emotions | None = None
    mask: Mask | None = None
    gaze: Gaze | None = None
    headPose: HeadPose | None = None
    mouthAttributes: MouthAttributes | None = None
    eyesAttributes: EyesAttributes | None = None
    liveness: Liveness | None = None
    deepfake: Deepfake | None = None


@dataclass
class FaceQualityFloatThreshold(BaseClass):
    """
    Face quality float threshold
    Attributes:
        min: minimal alloweb value
        max: maximum allowed value
    """

    min: float | None = None
    max: float | None = None


@dataclass
class FaceQualityIntThreshold(BaseClass):
    """
    Face quality int threshold
    Attributes:
        min: minimal alloweb value
        max: maximum allowed value
    """

    min: int | None = None
    max: int | None = None


@dataclass
class FaceQualityCheck(BaseClass):
    """
    Face quality check class
    Attributes:
        name: face quality check name
        objectValue: face quality checks estimation value
        thresholdValue: face quality check threshold
        result: face quality check result: 0 - fail, 1 - success
    """

    name: str
    objectValue: str | float | int
    thresholdValue: FaceQualityFloatThreshold | FaceQualityIntThreshold | list[str] | int
    result: int


@dataclass
class FaceQuality(BaseClass):
    """
    Face quality class

    Attributes:
        status: face quality checks status for all estimates
        checks: face quality container with all estimates
    """

    status: int
    checks: list[FaceQualityCheck]


@dataclass
class RawFaceDetection(BaseClass):
    """
    Raw face detection class

    Attributes:
        attributes: face detection attributes
        rect: face detection rect
        quality: face quality estimations
        landmarks5: array of 5 landmarks
        landmarks68: array of 68 landmarks
        faceQuality: face quality checks
    """

    attributes: RawFaceSampleAttributes | None = None
    rect: Rect | None = None
    quality: QualityEstimations | None = None
    landmarks5: Landmarks | None = None
    landmarks68: Landmarks | None = None
    faceQuality: FaceQuality | None = None


@dataclass
class RawBodyDetection(BaseClass):
    """
    Raw body detection class

    Attributes:
        attributes: body detection attributes
        rect: body detection rect
    """

    attributes: BodyExtraAttributes | None = None
    rect: Rect | None = None


@dataclass
class RawFace(BaseClass):
    """
    Event face class

    Attributes:
        faceId: created face id
        lists: attach result (list ids)
        eventId: event id
        userData: user data
        externalId: external id
        avatar: avatar for face
        url: location of face
    """

    faceId: str
    lists: list[str] | None = None
    eventId: str | None = None
    userData: str | None = None
    externalId: str | None = None
    avatar: str | None = None
    url: str | None = None


@dataclass
class RawFaceAttributes(BaseClass):
    """
    Raw face attributes class

    Attributes:
        attributeId: attribute id
        score: face descriptor quality score
        samples: resource for attributes
        basicAttributes: basic attributes
        descriptor: descriptor data
        url: location of attributes
    """

    attributeId: str | None = None
    score: float | None = None
    samples: list[str] | None = None
    basicAttributes: RawBasicAttributes | None = None
    descriptor: str | None = None
    url: str | None = None


@dataclass
class RawBodyAttributes(BaseClass):
    """
    Raw face attributes class

    Attributes:
        score: body descriptor quality score
        samples: resource for attributes
        descriptor: descriptor data
    """

    score: float | None = None
    samples: list[str] | None = None
    descriptor: str | None = None


@dataclass
class RawFaceSample(BaseClass):
    """
    Raw face sample class

    Attributes:
        sampleId: sample id
        detection: face detection
        url: location of the sample
    """

    sampleId: str | None
    detection: RawFaceDetection
    url: str | None = None


@dataclass
class RawBodySample(BaseClass):
    """
    Raw body sample class

    Attributes:
        sampleId: sample id
        detection: body detection
        url: location of the sample
    """

    sampleId: str | None
    detection: RawBodyDetection
    url: str | None = None


@dataclass
class RawDetectionSamples(BaseClass):
    """
    Detection samples class. The face and the body refer to the same human.

    Attributes:
        face: face sample
        body: body sample
    """

    face: RawFaceSample | None = None
    body: RawBodySample | None = None

    def asDict(self) -> dict:
        """
        Cast self into dict. Used by `ujson.dumps`.

        Returns:
            self as dict
        """
        return {"face": self.face.asDict() if self.face else None, "body": self.body.asDict() if self.body else None}


@dataclass
class RawDetections(BaseClass):
    """
    Raw event detections class

    Attributes:
        filename: source image filename
        samples: samples
        detectTime: time in format RFC 3339
        imageOrigin: the image url - source of the detection
        detectTs: user-defined timestamp relative to something, such as the start of a video
    """

    samples: RawDetectionSamples
    filename: str | None = None
    detectTime: str | None = None
    imageOrigin: str | None = None
    detectTs: float | None = None


@dataclass
class TopFaceMatchResult(BaseClass):
    """
    Face top match result class

    Attributes:
        label: label for matching candidates
        similarity: similarity to the most similar face or event
        faceId: face id
        externalId: external id
    """

    label: str
    similarity: float
    faceId: str
    externalId: str | None = None


@dataclass
class TopEventMatchResult(BaseClass):
    """
    Event top match result class

    Attributes:
        label: label for matching candidates
        similarity: similarity to the most similar face or event
        eventId: event id
        externalId: external id
    """

    label: str
    similarity: float
    eventId: str
    externalId: str | None = None


@dataclass
class DetectionResult(BaseClass):
    """
    Detection result class

    Attributes:
        rect: body detection rect
    """

    rect: Rect | None = None


@dataclass
class ExtractedDetections(BaseClass):
    """
    Class for extracted detections (face and body)

    Attributes:
        sampleId: sample id
        detection: face or body detection
        detectTime: time in format RFC 3339
        imageOrigin: the image url - source of the detection
        detectTs: user-defined timestamp relative to something, such as the start of a video
    """

    sampleId: str
    detection: DetectionResult
    detectTime: str
    imageOrigin: str | None = None
    detectTs: float | None = None


@dataclass
class RawFaceCandidate(BaseClass):
    """
    Raw face candidate class

    Attributes:
        faceId: face id
        createTime: create time
        externalId: external id
        userData: user data
        accountId: account id
        eventId: event id
        avatar: avatar for face
        lists: list of luna lists
    """

    faceId: str | None = None
    createTime: str | None = None
    externalId: str | None = None
    userData: str | None = None
    accountId: str | None = None
    eventId: str | None = None
    avatar: str | None = None
    lists: list[str] | None = None


@dataclass
class RawEventCandidate(BaseClass):
    """
    Raw event candidate class

    Attributes:
        eventId: event id
        streamId: stream id
        source: source
        handlerId: handler id
        externalId: external id
        userData: user data
        createTime: create time
        endTime: end time
        accountId: account id
        faceId: face id
        attachResult: list of attaching faces or bodies
        gender: gender, 0 - female, 1 - male
        age: human age
        emotion: predominant emotion
        mask: predominant mask
        ethnicGroup: ethnicity
        tags: list of tags
        location: information about a place where event is incidented
        trackId: event track id
        topMatch: top match result from match policy
        matchResult: match result for the event
        faceDetections: face detections
        bodyDetections: body detections
    """

    eventId: str | None = None
    accountId: str | None = None
    createTime: str | None = None
    endTime: str | None = None
    handlerId: str | None = None
    externalId: str | None = None
    userData: str | None = None
    streamId: str | None = None
    source: str | None = None
    faceId: str | None = None
    attachResult: list[str] | None = None
    gender: int | None = None
    age: int | None = None
    emotion: int | None = None
    mask: int | None = None
    ethnicGroup: int | None = None
    tags: list[str] | None = None
    location: Location | None = None
    trackId: str | None = None
    topMatch: TopFaceMatchResult | TopEventMatchResult | None = None
    matchResult: list[EventMatch] = None
    faceDetections: list[ExtractedDetections] | None = None
    bodyDetections: list[ExtractedDetections] | None = None
    bodyBasicAttributes: BodyBasicAttributes | None = None
    upperBody: BodyUpperAttributes | None = None
    accessories: BodyAccessories | None = None


@dataclass
class RawFaceMatchCandidates(BaseClass):
    """
    Raw face match candidates class

    Attributes:
        similarity: similarity
        face: candidate-face
    """

    similarity: float
    face: RawFaceCandidate


@dataclass
class RawEventMatchCandidates(BaseClass):
    """
    Raw event match candidates class

    Attributes:
        similarity: similarity
        event: candidate-event
    """

    similarity: float
    event: RawEventCandidate


@dataclass
class RawMatch(BaseClass):
    """
    Raw event match class

    Attributes:
        label: match label
        candidates: match candidates
    """

    label: str
    candidates: list[RawFaceMatchCandidates | RawEventMatchCandidates]


@dataclass
class FaceAggregateEstimationsAttributes(BaseClass):
    """
    Event aggregate estimations class

    Attributes:
        liveness: liveness detection
        deepfake: deepfake estimation
        emotions: face emotions
        mask: face mask
    """

    liveness: Liveness | None = None
    deepfake: Deepfake | None = None
    emotions: Emotions | None = None
    mask: Mask | None = None


@dataclass
class FaceAggregateEstimations(BaseClass):
    """
    Event face aggregate estimations class

    Attributes:
        attributes: face aggregate estimations attributes
    """

    attributes: FaceAggregateEstimationsAttributes


@dataclass
class AggregateEstimations(BaseClass):
    """
    Event aggregated estimations class

    Attributes:
        face: face aggregated estimations
        body: body aggregated estimations
    """

    face: FaceAggregateEstimations | None = None
    body: BodyAggregateEstimations | None = None


@dataclass
class RawEvent(BaseClass):
    """
    Raw event class

    Attributes:
        createTime: create time
        endTime: end time
        detections: list of body & face related detections
        faceAttributes: extracted face attributes from samples
        bodyAttributes: extracted body attributes from samples
        externalId: external id
        face: created face
        location: location
        matches: list with match result for the event
        source: source
        streamId: stream id
        tags: list of tags
        trackId: track id
        userData: user data
        aggregateEstimations: aggregate estimations
        meta: user-defined event metadata
    """

    detections: list[RawDetections] | None = None
    location: Location | None = None
    faceAttributes: RawFaceAttributes | None = None
    bodyAttributes: RawBodyAttributes | None = None
    externalId: str | None = None
    face: RawFace | None = None
    matches: list[RawMatch] | None = None
    source: str | None = None
    streamId: str | None = None
    tags: list[str] | None = None
    trackId: str | None = None
    userData: str | None = None
    createTime: str | None = None
    endTime: str | None = None
    aggregateEstimations: AggregateEstimations | None = None
    meta: dict | None = None


@dataclass
class FaceUpgradePolicy(BaseClass):
    """
    Face upgrade policy

    Attributes:
        fdScoreThreshold: face descriptor score threshold
        extractDescriptor: whether to extract face descriptor
        extractBasicAttributes: whether to extract basic attributes
        descriptorVersion: face descriptor version
    """

    fdScoreThreshold: float | None = None
    extractDescriptor: int | None = None
    extractBasicAttributes: int | None = None
    descriptorVersion: int | None = None


@dataclass
class BodyUpgradePolicy(BaseClass):
    """
    Body upgrade policy

    Attributes:
        descriptorVersion: body descriptor version
    """

    descriptorVersion: int | None = None


@dataclass
class FaceUpgradeData(BaseClass):
    """
    Face upgrade data

    Attributes:
        samples: face samples
        policy: face attribute upgrade policy
    """

    samples: list[str]
    policy: FaceUpgradePolicy | None = None


@dataclass
class BodyUpgradeData(BaseClass):
    """
    Body upgrade data

    Attributes:
        samples: body samples
        policy: body attribute upgrade policy
    """

    samples: list[str]
    policy: BodyUpgradePolicy | None = None


@dataclass
class FaceAttributeUpgrade(BaseClass):
    """
    Face attribute upgrade class

    Attributes:
        faceId: face id
        upgradeData: samples and policy to upgrade face attribute
    """

    faceId: str
    upgradeData: FaceUpgradeData


@dataclass
class EventAttributeUpgrade(BaseClass):
    """
    Event attribute upgrade class

    Attributes:
        eventId: event id
        faceUpgradeData: samples and policy to upgrade face attribute
        bodyUpgradeData: samples and policy to upgrade body attribute
    """

    eventId: str
    faceUpgradeData: FaceUpgradeData | None = None
    bodyUpgradeData: BodyUpgradeData | None = None


@dataclass
class StreamEventSourceData(BaseClass):
    """Stream event source data"""

    source: str | None = None
    streamId: str | None = None
    tags: list[str] | None = None
    userData: str | None = None
    externalId: str | None = None
    trackId: str | None = None
    meta: dict | None = None
    eventTime: str | None = None
    eventEndTime: str | None = None
    location: Location | None = None


@dataclass
class ImageOrigin(BaseClass):
    """Image origin"""

    body: bytes
    meta: dict | None = None


@dataclass
class BoundingBox(BaseClass):
    """
    Bounding box
    """

    x: int  # pylint: disable-msg=invalid-name
    y: int  # pylint: disable-msg=invalid-name
    width: int
    height: int


@dataclass
class AggregatedRawImage(BaseClass):
    """Aggregated raw image"""

    body: bytes
    faceBoundingBoxes: list[BoundingBox] | None = None
    bodyBoundingBoxes: list[BoundingBox] | None = None
    meta: dict | None = None


@dataclass
class NonAggregatedRawImage(BaseClass):
    """Non aggregated raw image"""

    body: bytes
    sourceData: StreamEventSourceData | None = None
    faceBoundingBoxes: list[BoundingBox] | None = None
    bodyBoundingBoxes: list[BoundingBox] | None = None
    meta: dict | None = None


@dataclass
class RawDetection(BaseClass):
    """Raw detection"""

    warp: bytes | None = None
    boundingBox: BoundingBox | None = None
    meta: dict | None = None


@dataclass
class AggregatedDetections(BaseClass):
    """Aggregated detections"""

    face: RawDetection | None = None
    body: RawDetection | None = None


@dataclass
class NonAggregatedDetections(BaseClass):
    """Aggregated detections"""

    sourceData: StreamEventSourceData | None = None
    face: RawDetection | None = None
    body: RawDetection | None = None


@dataclass
class StreamEventSourceNonAggregated(BaseClass):
    """Stream event source non aggregated"""

    source: NonAggregatedRawImage | list[NonAggregatedDetections]
    sourceType: Literal["raw_image", "detections"] = None

    filename: str | None = None
    detectTime: str | None = None
    detectTs: float | None = None
    imageOrigin: ImageOrigin | str | None = None


@dataclass
class StreamEventSourceAggregated(BaseClass):
    """Stream event source aggregated"""

    source: AggregatedRawImage | list[AggregatedDetections]
    sourceType: Literal["raw_image", "detections"] = None

    filename: str | None = None
    detectTime: str | None = None
    detectTs: float | None = None
    imageOrigin: ImageOrigin | str | None = None


@dataclass
class StreamEventsSource(BaseClass):
    """
    Stream events source
    Request with attributes aggregation fits with `sourceData` at top level and `StreamEventSourceAggregated`
    Request without attributes aggregation fits with `StreamEventSource` and `sourceData` at each `StreamEventSource`
    """

    sources: list[StreamEventSourceNonAggregated] | list[StreamEventSourceAggregated]
    aggregateAttributes: int
    sourceData: StreamEventSourceData | None = None
    useExifInfo: int | None = None
