"""
Luna-image-store client module
"""

from numbers import Real
from typing import Awaitable, List, Optional, Union

import aiohttp

from ..common.base_api import BaseHTTPAPI
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload


class StoreApi(BaseHTTPAPI):
    """
    Class for request to luna-image-store.

    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5020",
        api: Optional[int] = 1,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def createBucket(self, bucketName: Optional[str] = "", **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create bucket with bucketName.

        Args:
            bucketName: bucket's names.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"bucket": bucketName}
        return self.makeRequest(f"{self.baseUri}/buckets", "POST", queryParams=queries, **kwargs)

    def getBucket(self, bucketName: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get buckets with bucketName.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}", "GET", **kwargs)

    def headBucket(self, bucketName: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Checks whether bucket with bucketName exists.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}", "HEAD", **kwargs)

    def getBuckets(self, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get buckets with lunaRequestId.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        return self.makeRequest(f"{self.baseUri}/buckets", "GET", **kwargs)

    def postImage(
        self,
        imageInBytes: Union[bytes, bytearray],
        bucketName: str,
        accountId: Optional[str] = None,
        contentType: Optional[str] = "image/jpeg",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Post an image to bucket.

        Args:
            imageInBytes: byte's array (image)
            bucketName: bucket's name
            accountId: A str, account id
            contentType: content-type of image or image/jpeg for default

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        payload = RequestPayload(body=imageInBytes, headers={"Content-Type": contentType})
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}/images", "POST", body=payload, **kwargs)

    def deleteBucket(self, bucketName: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete bucket with bucketName.

        Args:
            bucketName: bucket's name

        Returns:
            In body of :class:`~.LunaResponse` will be None.

        """
        queries = {"bucket": bucketName}
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}", "DELETE", queryParams=queries, **kwargs)

    def putImage(
        self,
        imageInBytes: Union[bytes, bytearray],
        imageId: str,
        bucketName: str,
        accountId: Optional[str] = None,
        contentType: Optional[str] = "image/jpeg",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Put an image to bucket.

        Args:
            imageInBytes: byte's array (image)
            imageId: image id
            bucketName: bucket's name
            accountId: A str, account id
            contentType: content-type of image or image/jpeg for default

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        payload = RequestPayload(body=imageInBytes, headers={"Content-Type": contentType})
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}/images/{imageId}", "PUT", body=payload, **kwargs)

    def checkImage(
        self, bucketName: str, imageId: str, accountId: Optional[str] = None, withMeta: bool = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check image existence in bucket.

        Args:
            bucketName: bucket's name
            imageId: external image id
            accountId: A str, account id
            withMeta: whether to retrieve user-defined image metadata

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId, "with_meta": withMeta}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/images/{imageId}", "HEAD", queryParams=queries, **kwargs
        )

    def getImage(
        self, bucketName: str, imageId: str, accountId: Optional[str] = None, withMeta: bool = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get image from bucket.

        Args:
            bucketName: bucket's name
            imageId: external image id
            accountId: A str, account id
            withMeta: whether to retrieve user-defined image metadata

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId, "with_meta": withMeta}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/images/{imageId}",
            "GET",
            queryParams=queries,
            acceptEncoding="identity",
            **kwargs,
        )

    def deleteImage(
        self, bucketName: str, imageId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete image from bucket.

        Args:
            bucketName: bucket's name
            imageId: image id to delete
            accountId: A str, account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/images/{imageId}", "DELETE", queryParams=queries, **kwargs
        )

    def deleteImages(
        self, bucketName: str, imageIds: List[str], accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete images from bucket.

        Args:
            bucketName: bucket's name
            imageIds: images ids to delete
            accountId: A str, account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        payload = RequestPayload.buildJson(body={"images": imageIds})
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/images", "DELETE", queryParams=queries, body=payload, **kwargs
        )

    def postObject(
        self,
        objectBody: str,
        bucketName: str,
        accountId: Optional[str] = None,
        contentType: Optional[str] = "application/json",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Post object to bucket.

        Args:
            objectBody: object, available: text, json
            bucketName: bucket's name
            accountId: A str, account id
            contentType: content-type of object or application/json for default

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        payload = RequestPayload(body=objectBody, headers={"Content-Type": contentType})
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/buckets/{bucketName}/objects", "POST", body=payload, **kwargs)

    def putObject(
        self,
        objectBody: str,
        objectId: str,
        bucketName: str,
        accountId: Optional[str] = None,
        contentType: Optional[str] = "application/json",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Put object to bucket.

        Args:
            objectBody: object, available: text, json
            objectId: object id
            bucketName: bucket's name
            accountId: A str, account id
            contentType: content-type of object or application/json for default

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        payload = RequestPayload(body=objectBody, headers={"Content-Type": contentType})
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/objects/{objectId}",
            "PUT",
            body=payload,
            **kwargs,
        )

    def getObject(
        self,
        bucketName: str,
        objectId: str,
        accountId: Optional[str] = None,
        acceptType: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get object from bucket.

        Args:
            bucketName: bucket's name
            objectId: object id
            accountId: A str, account id
            acceptType: acceptable content-type of response body

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        headers = {"Accept": acceptType} if acceptType else {}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/objects/{objectId}",
            "GET",
            queryParams=queries,
            headers=headers,
            acceptEncoding="identity",
            accept=acceptType,
            **kwargs,
        )

    def checkObject(
        self,
        bucketName: str,
        objectId: str,
        accountId: Optional[str] = None,
        acceptType: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check object existence in bucket.

        Args:
            bucketName: bucket's name
            objectId: object id
            accountId: A str, account id
            acceptType: acceptable content-type of response body

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        headers = {"Accept": acceptType} if acceptType else {}
        queries = {"account_id": accountId}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/objects/{objectId}",
            "HEAD",
            queryParams=queries,
            headers=headers,
            **kwargs,
        )

    def deleteObject(
        self, bucketName: str, objectId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete object from bucket.

        Args:
            bucketName: bucket's name
            objectId: object id
            accountId: A str, account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/objects/{objectId}", "DELETE", queryParams=queries, **kwargs
        )

    def deleteObjects(
        self, bucketName: str, objectIds: List[str], accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete objects from bucket.

        Args:
            bucketName: bucket's name
            objectIds: objects ids to delete
            accountId: A str, account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with version.

        """
        queries = {"account_id": accountId}
        payload = RequestPayload.buildJson(body={"objects": objectIds})
        return self.makeRequest(
            f"{self.baseUri}/buckets/{bucketName}/objects", "DELETE", queryParams=queries, body=payload, **kwargs
        )
