"""
Module realize handlers mixin
"""

from typing import Any, Awaitable, Dict, List, Optional, Union

from luna3.common import functions
from luna3.common.http_objs import (
    BinaryDescriptorData,
    BinaryImage,
    Policies,
    SampleForDetection,
    UrlForDetection,
    VerifierPolicies,
)
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.handlers.http_objs import StreamEventsSource
from luna3.lunavl.http_objs import ApiRawEvent
from luna3.lunavl.mixins.base import MixinStub
from luna3.python_matcher.match_objects import BinaryReference


class HandlersMixin(MixinStub):
    """Handlers Mixin to work with handlers."""

    def createHandler(
        self,
        policies: Optional[Policies] = None,
        handlerType: Optional[int] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new handler. Handler determinate a list of rules for proccessing input images.
        You can set parameters for detecting faces, extracting attributes.
        You can set rules for matching attributes and creating faces from them.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createHandler

        Args:
            policies: object with policies information
            handlerType: handler type (0 - static, 1 - dynamic, 2 - lambda)
            description: user description

        Returns:
            LunaResponse: body contains json with new handler id and url.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy, StoragePolicy, LinkToListPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Filters, Policies, FacePolicy
        >>> from luna3.common.http_objs import MatchPolicyCandidates, ConditionalTagPolicy
        >>>
        >>> response = client.createHandler(handlerType=1, description="dynamic")
        >>> print(response.statusCode)
        201
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"external_url": "http://127.0.0.1:5000/6/handlers/...", "handler_id": ..., "url": ...}
        >>> listId = client.createList("strange_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(
        ...         estimateHeadPose=1,
        ...         estimateEmotions=1,
        ...         estimateGaze=1,
        ...         extractExif=1,
        ...         estimateLiveness=DetectPolicy.LivenessEstimationPolicy(1, 0.5, 0.5),
        ...         estimateDeepfake=DetectPolicy.DeepfakeEstimationPolicy(1, 0.5, 2),
        ...         rollThreshold=20,
        ...     ),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.7772),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="test_faces",
        ...             gender=1,
        ...             ethnicities=[1, 2],
        ...             limit=3,
        ...         )
        ...     ],
        ...     storagePolicy=StoragePolicy(
        ...         facePolicy=FacePolicy(
        ...             storeFace=1,
        ...             filters=Filters(ageGte=18, ageLt=40),
        ...             setSampleAsAvatar=1,
        ...             linkToListsPolicy=[LinkToListPolicy(listId=listId)],
        ...         )
        ...     ),
        ...     conditionalTagsPolicy=[ConditionalTagPolicy(tag="test_tag")]
        ... )
        >>> response = client.createHandler(description="persons", policies=policies)
        >>> print(response.statusCode)
        201
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"external_url": "http://127.0.0.1:5000/6/handlers/...", "handler_id": ..., "url": ...}

        """
        data = {}
        if description is not None:
            data["description"] = description
        if policies is not None:
            data["policies"] = policies.asDict()
        if handlerType is not None:
            data["handler_type"] = handlerType
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/handlers", "POST", body=payload, **kwargs)

    def getHandlers(
        self,
        description: Optional[str] = None,
        handlerType: Optional[int] = None,
        page: int = 1,
        pageSize: int = 10,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handlers by filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getHandlers

        Args:
            description: Find all handlers with description which "like" of this parameter
            handlerType: handler type (0 - static, 1 - dynamic, 2 - lambda)
            page: The number of page
            pageSize: The numbers of items on page
            accountIdFilter: account id to get handlers

        Returns:
            LunaResponse: body contains json with handlers info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy, StoragePolicy, LinkToListPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Filters, Policies, FacePolicy
        >>> from luna3.common.http_objs import MatchPolicyCandidates, ConditionalTagPolicy
        >>>
        >>> _ = client.createHandler(handlerType=1, description="dynamic", raiseError=True)
        >>> response = client.getHandlers(description="dynamic", handlerType=1)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json[0], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "description": "dynamic",
            "handler_id": ...,
            "handler_type": 1,
            "is_dynamic": ...,
            "last_update_time": ...,
            "policies": null
        }
        >>> listId = client.createList("strange_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(
        ...         estimateHeadPose=1,
        ...         estimateEmotions=1,
        ...         estimateGaze=1,
        ...         extractExif=1,
        ...         estimateLiveness=DetectPolicy.LivenessEstimationPolicy(1, 0.5, 0.5),
        ...         estimateDeepfake=DetectPolicy.DeepfakeEstimationPolicy(1, 0.5, 2),
        ...         rollThreshold=20,
        ...     ),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.7772),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="test_faces",
        ...             gender=1,
        ...             ethnicities=[1, 2],
        ...             limit=3,
        ...         )
        ...     ],
        ...     storagePolicy=StoragePolicy(
        ...         facePolicy=FacePolicy(
        ...             storeFace=1,
        ...             filters=Filters(ageGte=18, ageLt=40),
        ...             setSampleAsAvatar=1,
        ...             linkToListsPolicy=[LinkToListPolicy(listId=listId)],
        ...         )
        ...     ),
        ...     conditionalTagsPolicy=[ConditionalTagPolicy(tag="test_tag")],
        ... )
        >>> response = client.createHandler(description="test_persons", policies=policies, raiseError=True)
        >>> response = client.getHandlers(description="test_persons", handlerType=0)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json[0], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "description": "test_persons",
            "handler_id": ...,
            "handler_type": 0,
            "is_dynamic": ...,
            "last_update_time": ...,
            "policies": {
                "conditional_tags_policy": [
                    {
                        "filters": {},
                        "tag": "test_tag"
                    }
                ],
                "detect_policy": {
                    "body_attributes": {
                        "estimate_accessories": 0,
                        "estimate_basic_attributes": 0,
                        "estimate_lower_body": 0,
                        "estimate_upper_body": 0
                    },
                    "deepfake_states": null,
                    "detect_body": 0,
                    "detect_face": 1,
                    "detect_landmarks68": 0,
                    "estimate_deepfake": {
                        "estimate": 1,
                        "mode": 2,
                        "real_threshold": 0.5
                    },
                    "estimate_emotions": 1,
                    "estimate_eyes_attributes": 0,
                    "estimate_gaze": 1,
                    "estimate_glasses": 0,
                    "estimate_head_pose": 1,
                    "estimate_liveness": {
                        "estimate": 1,
                        "liveness_threshold": 0.5,
                        "quality_threshold": 0.5
                    },
                    "estimate_mask": 0,
                    "estimate_mouth_attributes": 0,
                    "estimate_people_count": {
                        "estimate": 0
                    },
                    "estimate_quality": 0,
                    "extract_exif": 1,
                    "face_quality": {
                        "checks": {...},
                        "estimate": 0,
                        "filter": 0
                    },
                    "liveness_states": null,
                    "mask_states": null,
                    "multiface_policy": 1,
                    "pitch_threshold": null,
                    "roll_threshold": 20,
                    "yaw_threshold": null
                },
                "extract_policy": {
                    "extract_basic_attributes": 1,
                    "extract_body_descriptor": 0,
                    "extract_face_descriptor": 1,
                    "fd_score_threshold": 0.7772
                },
                "match_policy": [
                    {
                        "candidates": {
                            "list_id": ...,
                            "origin": "faces"
                        },
                        "descriptor": {
                            "descriptor_type": "face"
                        },
                        "filters": {
                            "ethnicities": [
                                1,
                                2
                            ],
                            "gender": 1
                        },
                        "label": "test_faces",
                        "limit": 3
                    }
                ],
                "storage_policy": {
                    "attribute_policy": {
                        "filters": {},
                        "store_attribute": 0,
                        "ttl": 300
                    },
                    "body_sample_policy": {
                        "filters": {},
                        "store_sample": 1
                    },
                    "callbacks": [],
                    "event_policy": {
                        "filters": {},
                        "store_event": 1,
                        "wait_saving": 1
                    },
                    "face_policy": {
                        "filters": {
                            "age__gte": 18,
                            "age__lt": 40
                        },
                        "link_to_lists_policy": [
                            {
                                "filters": {},
                                "list_id": ...
                            }
                        ],
                        "set_sample_as_avatar": 1,
                        "store_face": 1
                    },
                    "face_sample_policy": {
                        "filters": {},
                        "store_sample": 1
                    },
                    "image_origin_policy": {
                        "filters": {},
                        "store_image": 0,
                        "use_external_references": 1
                    },
                    "notification_policy": {
                        "filters": {},
                        "send_notification": 1
                    }
                }
            }
        }

        """
        queries = {
            "description": description,
            "page": page,
            "page_size": pageSize,
            "handler_type": handlerType,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/handlers", "GET", queryParams=queries, **kwargs)

    def getHandlersCount(
        self,
        description: Optional[str] = None,
        handlerType: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handlers count by filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getHandlerCount

        Args:
            description: Find all handlers with description which "like" of this parameter
            handlerType: handler type (0 - static, 1 - dynamic, 2 - lambda)
            accountIdFilter: account id to get handlers count

        Returns:
            LunaResponse: body contains json with handlers count

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> response = client.createHandler(handlerType=1, description="dynamic", raiseError=True)
        >>> response = client.getHandlersCount(description="dynamic", handlerType=1)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"handlers_count": ...}

        """
        queries = {"description": description, "handler_type": handlerType, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/handlers/count", "GET", queryParams=queries, **kwargs)

    def getHandlerById(
        self, handlerId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get handler by its id

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getHandler

        Args:
            handlerId: handler id
            accountIdFilter: account id to get handler

        Returns:
            LunaResponse: body contains json with handler info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> handlerId = client.createHandler(handlerType=1, description="dynamic", raiseError=True).json["handler_id"]
        >>> response = client.getHandlerById(handlerId=handlerId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "description": "dynamic",
            "handler_id": ...,
            "handler_type": 1,
            "is_dynamic": ...,
            "last_update_time": ...,
            "policies": null
        }

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "GET", queryParams=queries, **kwargs)

    def updateHandler(
        self,
        handlerId: str,
        policies: Optional[Policies] = None,
        handlerType: Optional[int] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update a handler. You can not update a part of handler.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/putHandler

        Args:
            handlerId: handler id
            policies: object with policies information
            handlerType: handler type (0 - static, 1 - dynamic, 2 - lambda)
            description: user description

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> handlerId = client.createHandler(handlerType=1, description="dynamic", raiseError=True).json['handler_id']
        >>> response = client.updateHandler(
        ...    handlerId=handlerId, handlerType=1, description="test_persons_3"
        ... )
        >>> print(response.statusCode)
        204

        """
        data = {}
        if description is not None:
            data["description"] = description
        if policies is not None:
            data["policies"] = policies.asDict()
        if handlerType is not None:
            data["handler_type"] = handlerType
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "PUT", body=payload, **kwargs)

    def validateHandler(
        self,
        policies: Policies,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Validate handler policies.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/validatePolicies

        Args:
            policies: object with policies information

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy, StoragePolicy, LinkToListPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Filters, Policies, FacePolicy
        >>> from luna3.common.http_objs import MatchPolicyCandidates, ConditionalTagPolicy
        >>>
        >>> listId = client.createList("strange_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(
        ...         estimateHeadPose=1,
        ...         estimateEmotions=1,
        ...         estimateGaze=1,
        ...         extractExif=1,
        ...         estimateLiveness=DetectPolicy.LivenessEstimationPolicy(1, 0.5, 0.5),
        ...         estimateDeepfake=DetectPolicy.DeepfakeEstimationPolicy(1, 0.5, 2),
        ...         rollThreshold=20,
        ...     ),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.7772),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="test_faces",
        ...             gender=1,
        ...             ethnicities=[1, 2],
        ...             limit=3,
        ...         )
        ...     ],
        ...     storagePolicy=StoragePolicy(
        ...         facePolicy=FacePolicy(
        ...             storeFace=1,
        ...             filters=Filters(ageGte=18, ageLt=40),
        ...             setSampleAsAvatar=1,
        ...             linkToListsPolicy=[LinkToListPolicy(listId=listId)],
        ...         )
        ...     ),
        ...     conditionalTagsPolicy=[ConditionalTagPolicy(tag="test_tag")],
        ... )
        >>> response = client.validateHandler(policies=policies)
        >>> print(response.statusCode)
        204

        """
        payload = RequestPayload.buildJson(body={"policies": policies.asDict()})
        return self.makeRequest(f"{self.baseUri}/handlers/validator", "POST", body=payload, **kwargs)

    def checkHandlerById(
        self, handlerId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check whether handler exists by its id

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/checkHandler

        Args:
            handlerId: handler id
            accountIdFilter: account id to check handler

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> handlerId = client.createHandler(handlerType=1, description="dynamic", raiseError=True).json["handler_id"]
        >>> response = client.checkHandlerById(handlerId=handlerId)
        >>> print(response.statusCode)
        200

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "HEAD", queryParams=queries, **kwargs)

    def deleteHandlerById(self, handlerId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove handler by id

         See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/removeHandler

         Args:
             handlerId: handler id

         Returns:
             LunaResponse: not contains response body.

         >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
         >>>
         >>> handlerId = client.createHandler(handlerType=1, description="dynamic", raiseError=True).json["handler_id"]
         >>> response = client.deleteHandlerById(handlerId=handlerId)
         >>> print(response.statusCode)
         204

        """
        return self.makeRequest(f"{self.baseUri}/handlers/{handlerId}", "DELETE", **kwargs)

    def emitEvents(
        self,
        handlerId: str,
        inputData: Union[BinaryImage, List[BinaryImage], List[UrlForDetection], List[str], List[SampleForDetection]],
        policies: Optional[Policies] = None,
        externalId: Optional[str] = None,
        userData: Optional[str] = None,
        aggregateAttributes: Optional[int] = None,
        source: Optional[str] = None,
        streamId: Optional[str] = None,
        city: Optional[str] = None,
        street: Optional[str] = None,
        district: Optional[str] = None,
        area: Optional[str] = None,
        houseNumber: Optional[str] = None,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        tags: Optional[str] = None,
        trackId: Optional[str] = None,
        imageType: int = None,
        lunaEventTime: Optional[str] = None,
        lunaEventEndTime: Optional[str] = None,
        useExifInfo: Optional[int] = None,
        meta: Optional[dict] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Emit input events.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/generateEvents

        Args:
            handlerId: handler id
            inputData: one of:
                        - Single image or images list for extraction (supported formats "jpeg, png, tif, ppm, bmp").
                        - list of urls with images
                        - list of images' ids
            policies: object with policies information for dynamic handler
            externalId: external id
            userData: user data
            aggregateAttributes: aggregate all attributes (descriptot, age, gender) or not (0 or 1)
            source: source
            streamId: stream id
            city: event city
            street: event street
            area: event area
            district: event district
            houseNumber: event house number
            latitude: event latitude in degrees
            longitude: event longitude in degrees
            tags: tags
            trackId: event track id
            imageType: Image type (0 - raw image, 1 - face warped image, 2 - body warped image).
            lunaEventTime: user defined event create time
            lunaEventEndTime: user defined event end time
            useExifInfo: whether to use exif info for auto orientation
            meta: user-defined event json metadata

        Returns:
            LunaResponse: body contains json with created event info.

        >>> import json
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Policies
        >>> from luna3.common.http_objs import MatchPolicyCandidates
        >>> from testData.resources import FACE_WARP as image
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> listId = client.createList(userData="happy_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(estimateEmotions=1, estimateMask=1),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.0,),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="happy",
        ...             limit=3,
        ...             threshold=0.5,
        ...             gender=0,
        ...         )
        ...     ],
        ... )
        >>> handlerId = client.createHandler(
        ...     description="test_handler", policies=policies, raiseError=True
        ... ).json["handler_id"]
        >>> response = client.emitEvents(
        ...     handlerId,
        ...     inputData=[image],
        ...     city="Kazan",
        ...     street="Saban",
        ...     houseNumber="5",
        ...     tags="test_tag",
        ...     imageType=1,
        ...     useExifInfo=1,
        ...     meta={"foo": "bar"},
        ... )
        >>> print(response.statusCode)
        201
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "events": [
                {
                    "aggregate_estimations": {
                        "body": {
                            "attributes": {}
                        },
                        "face": {
                            "attributes": {
                                "emotions": {
                                    "estimations": {
                                        "anger": ...,
                                        "disgust": ...,
                                        "fear": ...,
                                        "happiness": ...,
                                        "neutral": ...,
                                        "sadness": ...,
                                        "surprise": ...
                                    },
                                    "predominant_emotion": "neutral"
                                },
                                "mask": {
                                    "estimations": {
                                        "medical_mask": ...,
                                        "missing": ...,
                                        "occluded": ...
                                    },
                                    "face_occlusion": {
                                        "estimations": {
                                            "chin": ...,
                                            "clear": ...,
                                            "correct": ...,
                                            "full": ...,
                                            "mouth": ...,
                                            "partially": ...
                                        },
                                        "predominant_occlusion": "clear"
                                    },
                                    "predominant_mask": "missing"
                                }
                            }
                        }
                    },
                    "body_attributes": null,
                    "detections": [
                        {
                            "detect_time": ...,
                            "detect_ts": null,
                            "filename": "Claudia_Schiffer_warp.jpg",
                            "image_origin": null,
                            "samples": {
                                "body": null,
                                "face": {
                                    "detection": {
                                        "attributes": {
                                            "emotions": {
                                                "estimations": {
                                                    "anger": ...,
                                                    "disgust": ...,
                                                    "fear": ...,
                                                    "happiness": ...,
                                                    "neutral": ...,
                                                    "sadness": ...,
                                                    "surprise": ...
                                                },
                                                "predominant_emotion": "neutral"
                                            },
                                            "mask": {
                                                "estimations": {
                                                    "medical_mask": ...,
                                                    "missing": ...,
                                                    "occluded": ...
                                                },
                                                "face_occlusion": {
                                                    "estimations": {
                                                        "chin": ...,
                                                        "clear": ...,
                                                        "correct": ...,
                                                        "full": ...,
                                                        "mouth": ...,
                                                        "partially": ...
                                                    },
                                                    "predominant_occlusion": "clear"
                                                },
                                                "predominant_mask": "missing"
                                            }
                                        }
                                    },
                                    "sample_id": ...,
                                    "url": "/6/samples/faces/..."
                                }
                            }
                        }
                    ],
                    "event_id": ...,
                    "external_id": "",
                    "face": null,
                    "face_attributes": {
                        "attribute_id": null,
                        "basic_attributes": {
                            "age": ...,
                            "ethnicities": {
                                "estimations": {
                                    "african_american": ...,
                                    "asian": ...,
                                    "caucasian": ...,
                                    "indian": ...
                                },
                                "predominant_ethnicity": "caucasian"
                            },
                            "gender": 0
                        },
                        "samples": [...],
                        "score": ...,
                        "url": null
                    },
                    "location": {
                        "city": "Kazan",
                        "house_number": "5",
                        "street": "Saban"
                    },
                    "matches": [
                        {
                            "candidates": [],
                            "label": "happy"
                        }
                    ],
                    "meta": {
                        "foo": "bar"
                    },
                    "source": null,
                    "stream_id": null,
                    "tags": [
                        "test_tag"
                    ],
                    "track_id": null,
                    "url": "/6/events/...",
                    "user_data": ""
                }
            ],
            "filtered_detections": {
                "face_detections": []
            },
            "images": [
                {
                    "error": {
                        "desc": "Success",
                        "detail": "Success",
                        "error_code": 0,
                        "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-0"
                    },
                    "filename": "Claudia_Schiffer_warp.jpg",
                    "image_estimations": {},
                    "status": 1
                }
            ]
        }
        """
        queries = {
            "external_id": externalId,
            "user_data": userData,
            "aggregate_attributes": aggregateAttributes,
            "source": source,
            "tags": tags,
            "stream_id": streamId,
            "image_type": imageType,
            "city": city,
            "house_number": houseNumber,
            "street": street,
            "district": district,
            "area": area,
            "latitude": latitude,
            "longitude": longitude,
            "track_id": trackId,
            "use_exif_info": useExifInfo,
        }
        headers = {}
        if lunaEventTime is not None:
            headers["Luna-Event-Time"] = lunaEventTime
        if lunaEventEndTime is not None:
            headers["Luna-Event-End-Time"] = lunaEventEndTime
        payload = functions.createPayloadWithImages(inputData, headers, self.multipartBoundary, policies, meta=meta)
        return self.makeRequest(
            f"{self.baseUri}/handlers/{handlerId}/events",
            "POST",
            queryParams=queries,
            body=payload,
            **kwargs,
        )

    def emitStreamEvents(
        self,
        handlerId: str,
        inputData: StreamEventsSource,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Emit beta events.

        See
          https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/generateStreamEvents

        Args:
            handlerId: handler id
            inputData: structure with stream event(s) source

        Returns:
            LunaResponse: body contains json with created event info.

        >>> import json
        >>> from luna3.common.http_objs import DetectPolicy, ExtractPolicy
        >>> from luna3.common.http_objs import MatchPolicy, Policies
        >>> from luna3.common.http_objs import MatchPolicyCandidates
        >>> from luna3.handlers.http_objs import (
        ...     NonAggregatedRawImage,
        ...     StreamEventSourceNonAggregated,
        ...     StreamEventsSource,
        ...     StreamEventSourceData,
        ...     )
        >>> from testData.resources import FACE_IMAGE
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> listId = client.createList(userData="happy_persons", raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(estimateEmotions=1, estimateMask=1),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.0,),
        ...     matchPolicy=[
        ...         MatchPolicy(
        ...             candidates=MatchPolicyCandidates(origin="faces", listId=listId),
        ...             label="happy",
        ...             limit=3,
        ...             threshold=0.5,
        ...             gender=0,
        ...         )
        ...     ],
        ... )
        >>> handlerId = client.createHandler(
        ...     description="test_handler", policies=policies, raiseError=True
        ... ).json["handler_id"]
        >>> response = client.emitStreamEvents(
        ...     handlerId,
        ...     inputData=StreamEventsSource(
        ...         aggregateAttributes=0,
        ...         sources=[
        ...             StreamEventSourceNonAggregated(
        ...                 sourceType="raw_image", source=NonAggregatedRawImage(body=FACE_IMAGE.body)
        ...             )
        ...         ],
        ...   )
        ... )
        >>> print(response.statusCode)
        201
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "events": [
                {
                    "aggregate_estimations": {
                        "body": {
                            "attributes": {}
                        },
                        "face": {
                            "attributes": {
                                "emotions": {
                                    "estimations": {
                                        "anger": ...,
                                        "disgust": ...,
                                        "fear": ...,
                                        "happiness": ...,
                                        "neutral": ...,
                                        "sadness": ...,
                                        "surprise": ...
                                    },
                                    "predominant_emotion": "neutral"
                                },
                                "mask": {
                                    "estimations": {
                                        "medical_mask": ...,
                                        "missing": ...,
                                        "occluded": ...
                                    },
                                    "face_occlusion": {
                                        "estimations": {
                                            "chin": ...,
                                            "clear": ...,
                                            "correct": ...,
                                            "full": ...,
                                            "mouth": ...,
                                            "partially": ...
                                        },
                                        "predominant_occlusion": "clear"
                                    },
                                    "predominant_mask": "missing"
                                }
                            }
                        }
                    },
                    "body_attributes": null,
                    "detections": [
                        {
                            "detect_time": ...,
                            "detect_ts": null,
                            "filename": "raw_image",
                            "image_origin": null,
                            "samples": {
                                "body": null,
                                "face": {
                                    "detection": {
                                        "attributes": {
                                            "emotions": {
                                                "estimations": {
                                                    "anger": ...,
                                                    "disgust": ...,
                                                    "fear": ...,
                                                    "happiness": ...,
                                                    "neutral": ...,
                                                    "sadness": ...,
                                                    "surprise": ...
                                                },
                                                "predominant_emotion": "neutral"
                                            },
                                            "mask": {
                                                "estimations": {
                                                    "medical_mask": ...,
                                                    "missing": ...,
                                                    "occluded": ...
                                                },
                                                "face_occlusion": {
                                                    "estimations": {
                                                        "chin": ...,
                                                        "clear": ...,
                                                        "correct": ...,
                                                        "full": ...,
                                                        "mouth": ...,
                                                        "partially": ...
                                                    },
                                                    "predominant_occlusion": "clear"
                                                },
                                                "predominant_mask": "missing"
                                            }
                                        },
                                        "landmarks5": [
                                            [
                                                ...,
                                                ...
                                            ],
                                            [
                                                ...,
                                                ...
                                            ],
                                            [
                                                ...,
                                                ...
                                            ],
                                            [
                                                ...,
                                                ...
                                            ],
                                            [
                                                ...,
                                                ...
                                            ]
                                        ],
                                        "rect": {
                                            "height": ...,
                                            "width": ...,
                                            "x": ...,
                                            "y": ...
                                        }
                                    },
                                    "sample_id": ...,
                                    "url": ...
                                }
                            }
                        }
                    ],
                    "event_id": ...,
                    "external_id": ...,
                    "face": null,
                    "face_attributes": {
                        "attribute_id": null,
                        "basic_attributes": {
                            "age": ...,
                            "ethnicities": {
                                "estimations": {
                                    "african_american": ...,
                                    "asian": ...,
                                    "caucasian": ...,
                                    "indian": ...
                                },
                                "predominant_ethnicity": "caucasian"
                            },
                            "gender": ...
                        },
                        "samples": [
                            ...
                        ],
                        "score": ...,
                        "url": null
                    },
                    "location": {},
                    "matches": [
                        {
                            "candidates": [],
                            "label": "happy"
                        }
                    ],
                    "meta": null,
                    "source": null,
                    "stream_id": null,
                    "tags": [],
                    "track_id": null,
                    "url": ...,
                    "user_data": ...
                }
            ],
            "filtered_detections": {
                "face_detections": []
            },
            "images": [
                {
                    "error": {
                        "desc": "Success",
                        "detail": "Success",
                        "error_code": 0,
                        "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-0"
                    },
                    "filename": "raw_image",
                    "image_estimations": {},
                    "status": 1
                }
            ]
        }
        """
        payload = RequestPayload.buildMsgpack(body=inputData.asDict())
        return self.makeRequest(
            f"{self.baseUri}/handlers/{handlerId}/stream_events",
            "POST",
            body=payload,
            **kwargs,
        )

    def saveEvent(
        self, handlerId: str, event: Union[ApiRawEvent, Dict[str, Any]], waitSaving: Optional[int] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Save raw event.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/saveEvent

        Args:
            handlerId: handler id
            event: raw event
            waitSaving: whether to wait for event saving

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> handlerId = client.createHandler(
        ...     handlerType=1, description="dynamic_handler", raiseError=True
        ... ).json["handler_id"]
        >>> event = {
        ...     "face_attributes": {
        ...         "basic_attributes": {
        ...             "age": 29,
        ...             "gender": 1,
        ...             "ethnicities": {
        ...                 "predominant_ethnicity": "asian",
        ...                 "estimations": {
        ...                     "african_american": 1.92238889737406e-12,
        ...                     "asian": 0.954671621322632,
        ...                     "caucasian": 0.045328326523304,
        ...                     "indian": 7.65100649502415e-10,
        ...                 },
        ...             },
        ...         }
        ...     },
        ...     "user_data": "event_user_data",
        ... }
        >>> response = client.saveEvent(handlerId=handlerId, event=event, waitSaving=1)
        >>> print(response.statusCode)
        201

        """
        queries = {"wait_saving": waitSaving}
        payload = RequestPayload.buildJson(body=event.asDict() if isinstance(event, ApiRawEvent) else event)
        return self.makeRequest(
            f"{self.baseUri}/handlers/{handlerId}/events/raw", "POST", queryParams=queries, body=payload, **kwargs
        )

    def createVerifier(
        self, policies: Optional[VerifierPolicies] = None, description: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new verifier. Verifier determinate a list of rules for processing & verification of input images.
        You can set parameters for detecting faces, extracting attributes.
        You can set rules for matching attributes and creating faces from them.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createVerifier

        Args:
            policies: object with policies information
            description: user description

        Returns:
            LunaResponse: body contains json with new verifier id and url

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> response = client.createVerifier(description="Happy_persons", policies=policies)
        >>> print(response.statusCode)
        201
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"external_url": "http://127.0.0.1:5000/6/verifiers/...", "url": ..., "verifier_id": ...}

        """
        data = {}
        if description is not None:
            data["description"] = description
        if policies is not None:
            data["policies"] = policies.asDict()
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/verifiers", "POST", body=payload, **kwargs)

    def getVerifier(
        self, verifierId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get verifier

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getVerifier

        Args:
            verifierId: verifier id
            accountIdFilter: account id to get verifier
        Returns:
            LunaResponse: body contains json with verifier info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="Happy_persons", policies=policies, raiseError=True
        ... ).json["verifier_id"]
        >>> response = client.getVerifier(verifierId=verifierId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "description": "Happy_persons",
            "last_update_time": ...,
            "policies": {
                "detect_policy": {
                    "deepfake_states": null,
                    "detect_landmarks68": 0,
                    "estimate_deepfake": {
                        "estimate": 0,
                        "mode": 2,
                        "real_threshold": 0.5
                    },
                    "estimate_emotions": 1,
                    "estimate_eyes_attributes": 0,
                    "estimate_gaze": 0,
                    "estimate_glasses": 0,
                    "estimate_head_pose": 0,
                    "estimate_liveness": {
                        "estimate": 0,
                        "liveness_threshold": 0.5,
                        "quality_threshold": 0.5
                    },
                    "estimate_mask": 0,
                    "estimate_mouth_attributes": 0,
                    "estimate_quality": 0,
                    "extract_exif": 0,
                    "face_quality": {
                        "checks": {
                            "aspect_ratio": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.8,
                                    "min": 0.74
                                }
                            },
                            "background_lightness": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.2
                                }
                            },
                            "background_uniformity": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.5
                                }
                            },
                            "blurriness_quality": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.61
                                }
                            },
                            "dark_quality": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.5
                                }
                            },
                            "dynamic_range": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.5
                                }
                            },
                            "eye_distance": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 90.0
                                }
                            },
                            "eyebrows_state": {
                                "estimate": 1,
                                "threshold": [
                                    "neutral"
                                ]
                            },
                            "face_color_type": {
                                "estimate": 1,
                                "threshold": [
                                    "color"
                                ]
                            },
                            "face_height": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 180
                                }
                            },
                            "face_width": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 180
                                }
                            },
                            "gaze_pitch": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 5.0,
                                    "min": -5.0
                                }
                            },
                            "gaze_yaw": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 5.0,
                                    "min": -5.0
                                }
                            },
                            "glasses": {
                                "estimate": 1,
                                "threshold": [
                                    "no_glasses",
                                    "eyeglasses"
                                ]
                            },
                            "head_height": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.9,
                                    "min": 0.6
                                }
                            },
                            "head_horizontal_center": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.55,
                                    "min": 0.45
                                }
                            },
                            "head_pitch": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 5.0,
                                    "min": -5.0
                                }
                            },
                            "head_roll": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 8.0,
                                    "min": -8.0
                                }
                            },
                            "head_vertical_center": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.5,
                                    "min": 0.3
                                }
                            },
                            "head_width": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.75,
                                    "min": 0.5
                                }
                            },
                            "head_yaw": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 5.0,
                                    "min": -5.0
                                }
                            },
                            "headwear_type": {
                                "estimate": 1,
                                "threshold": [
                                    "none"
                                ]
                            },
                            "illumination_quality": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.3
                                }
                            },
                            "illumination_uniformity": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.5
                                }
                            },
                            "image_format": {
                                "estimate": 1,
                                "threshold": [
                                    "JPEG",
                                    "JPEG2000",
                                    "PNG"
                                ]
                            },
                            "image_height": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1080,
                                    "min": 180
                                }
                            },
                            "image_size": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 2097152,
                                    "min": 5120
                                }
                            },
                            "image_width": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1920,
                                    "min": 180
                                }
                            },
                            "indent_left": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 20
                                }
                            },
                            "indent_lower": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 20
                                }
                            },
                            "indent_right": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 20
                                }
                            },
                            "indent_upper": {
                                "estimate": 1,
                                "threshold": {
                                    "min": 20
                                }
                            },
                            "left_eye": {
                                "estimate": 1,
                                "threshold": [
                                    "open"
                                ]
                            },
                            "light_quality": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.57
                                }
                            },
                            "mouth_occluded": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.5,
                                    "min": 0.0
                                }
                            },
                            "mouth_open": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.5,
                                    "min": 0.0
                                }
                            },
                            "mouth_smiling": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 0.5,
                                    "min": 0.0
                                }
                            },
                            "natural_light": {
                                "estimate": 1,
                                "threshold": 1
                            },
                            "radial_distortion": {
                                "estimate": 1,
                                "threshold": 0
                            },
                            "red_eyes": {
                                "estimate": 1,
                                "threshold": 0
                            },
                            "right_eye": {
                                "estimate": 1,
                                "threshold": [
                                    "open"
                                ]
                            },
                            "shoulders_position": {
                                "estimate": 1,
                                "threshold": [
                                    "parallel"
                                ]
                            },
                            "smile_properties": {
                                "estimate": 1,
                                "threshold": [
                                    "none"
                                ]
                            },
                            "specularity_quality": {
                                "estimate": 1,
                                "threshold": {
                                    "max": 1.0,
                                    "min": 0.3
                                }
                            }
                        },
                        "estimate": 0,
                        "filter": 0
                    },
                    "liveness_states": null,
                    "mask_states": null,
                    "multiface_policy": 1,
                    "pitch_threshold": null,
                    "roll_threshold": null,
                    "yaw_threshold": null
                },
                "extract_policy": {
                    "extract_basic_attributes": 1,
                    "fd_score_threshold": 0.0
                },
                "storage_policy": {
                    "attribute_policy": {
                        "store_attribute": 1
                    },
                    "face_sample_policy": {
                        "store_sample": 1
                    }
                },
                "verification_threshold": 0.5
            },
            "verifier_id": ...,
            "version": 0
        }

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/verifiers/{verifierId}", "GET", queryParams=queries, **kwargs)

    def putVerifier(
        self, verifierId: str, policies: Optional[VerifierPolicies] = None, description: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Replace verifier

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/putVerifier

        Args:
            verifierId: verifier id
            policies: object with policies information
            description: user description

        Returns:
            LunaResponse: body contains json with verifier current version.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="Happy_persons", policies=policies, raiseError=True
        ... ).json["verifier_id"]
        >>> newPolicies = policies
        >>> newPolicies.verificationThreshold = 0.9
        >>> response = client.putVerifier(verifierId=verifierId, policies=newPolicies)
        >>> print(response.statusCode)
        200
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'version': 1}

        """
        data = {}
        if description is not None:
            data["description"] = description
        if policies is not None:
            data["policies"] = policies.asDict()
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/verifiers/{verifierId}", "PUT", body=payload, **kwargs)

    def checkVerifier(
        self, verifierId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check verifier existence

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/checkVerifier

        Args:
            verifierId: verifier id
            accountIdFilter: account id to check verifier

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="Happy_persons", policies=policies, raiseError=True
        ... ).json["verifier_id"]
        >>> response = client.checkVerifier(verifierId=verifierId)
        >>> print(response.statusCode)
        200

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/verifiers/{verifierId}", "HEAD", queryParams=queries, **kwargs)

    def deleteVerifier(self, verifierId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove verifier

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/removeVerifier

        Args:
            verifierId: verifier id

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="Happy_persons", policies=policies, raiseError=True
        ... ).json["verifier_id"]
        >>> response = client.deleteVerifier(verifierId=verifierId)
        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/verifiers/{verifierId}", "DELETE", **kwargs)

    def getVerifiersCount(
        self, description: Optional[str] = None, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Count verifiers that satisfy the filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/countVerifiers

        Args:
            description: verifier description ("like" parameter)
            accountIdFilter: account id to count verifiers

        Returns:
            LunaResponse: body contains json with verifier count

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> response = client.createVerifier(description="Happy_persons", policies=policies, raiseError=True)
        >>> response = client.getVerifiersCount(description="Happy_persons")
        >>> print(response.statusCode)
        200
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'verifiers_count': ...}

        """
        queries = {"description": description, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/verifiers/count", "GET", queryParams=queries, **kwargs)

    def getVerifiers(
        self,
        description: Optional[str] = None,
        page: int = 1,
        pageSize: int = 10,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get verifiers by filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getVerifiers

        Args:
            description: verifier description ("like" parameter)
            page: The number of page
            pageSize: The numbers of items on page
            accountIdFilter: account id to get verifiers

        Returns:
            LunaResponse: body contains json with verifiers info

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> response = client.createVerifier(description="Happy_persons", policies=policies, raiseError=True)
        >>> response = client.getVerifiers(description="Happy_persons", pageSize=1)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4)) # doctest: +ELLIPSIS
        [
            {
                "account_id": ...,
                "create_time": ...,
                "description": "Happy_persons",
                "last_update_time": ...,
                "policies": {
                    "detect_policy": {
                        "deepfake_states": null,
                        "detect_landmarks68": 0,
                        "estimate_deepfake": {
                            "estimate": 0,
                            "mode": 2,
                            "real_threshold": 0.5
                        },
                        "estimate_emotions": 1,
                        "estimate_eyes_attributes": 0,
                        "estimate_gaze": 0,
                        "estimate_glasses": 0,
                        "estimate_head_pose": 0,
                        "estimate_liveness": {
                            "estimate": 0,
                            "liveness_threshold": 0.5,
                            "quality_threshold": 0.5
                        },
                        "estimate_mask": 0,
                        "estimate_mouth_attributes": 0,
                        "estimate_quality": 0,
                        "extract_exif": 0,
                        "face_quality": {
                            "checks": {
                                "aspect_ratio": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.8,
                                        "min": 0.74
                                    }
                                },
                                "background_lightness": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.2
                                    }
                                },
                                "background_uniformity": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.5
                                    }
                                },
                                "blurriness_quality": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.61
                                    }
                                },
                                "dark_quality": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.5
                                    }
                                },
                                "dynamic_range": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.5
                                    }
                                },
                                "eye_distance": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 90.0
                                    }
                                },
                                "eyebrows_state": {
                                    "estimate": 1,
                                    "threshold": [
                                        "neutral"
                                    ]
                                },
                                "face_color_type": {
                                    "estimate": 1,
                                    "threshold": [
                                        "color"
                                    ]
                                },
                                "face_height": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 180
                                    }
                                },
                                "face_width": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 180
                                    }
                                },
                                "gaze_pitch": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 5.0,
                                        "min": -5.0
                                    }
                                },
                                "gaze_yaw": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 5.0,
                                        "min": -5.0
                                    }
                                },
                                "glasses": {
                                    "estimate": 1,
                                    "threshold": [
                                        "no_glasses",
                                        "eyeglasses"
                                    ]
                                },
                                "head_height": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.9,
                                        "min": 0.6
                                    }
                                },
                                "head_horizontal_center": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.55,
                                        "min": 0.45
                                    }
                                },
                                "head_pitch": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 5.0,
                                        "min": -5.0
                                    }
                                },
                                "head_roll": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 8.0,
                                        "min": -8.0
                                    }
                                },
                                "head_vertical_center": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.5,
                                        "min": 0.3
                                    }
                                },
                                "head_width": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.75,
                                        "min": 0.5
                                    }
                                },
                                "head_yaw": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 5.0,
                                        "min": -5.0
                                    }
                                },
                                "headwear_type": {
                                    "estimate": 1,
                                    "threshold": [
                                        "none"
                                    ]
                                },
                                "illumination_quality": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.3
                                    }
                                },
                                "illumination_uniformity": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.5
                                    }
                                },
                                "image_format": {
                                    "estimate": 1,
                                    "threshold": [
                                        "JPEG",
                                        "JPEG2000",
                                        "PNG"
                                    ]
                                },
                                "image_height": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1080,
                                        "min": 180
                                    }
                                },
                                "image_size": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 2097152,
                                        "min": 5120
                                    }
                                },
                                "image_width": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1920,
                                        "min": 180
                                    }
                                },
                                "indent_left": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 20
                                    }
                                },
                                "indent_lower": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 20
                                    }
                                },
                                "indent_right": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 20
                                    }
                                },
                                "indent_upper": {
                                    "estimate": 1,
                                    "threshold": {
                                        "min": 20
                                    }
                                },
                                "left_eye": {
                                    "estimate": 1,
                                    "threshold": [
                                        "open"
                                    ]
                                },
                                "light_quality": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.57
                                    }
                                },
                                "mouth_occluded": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.5,
                                        "min": 0.0
                                    }
                                },
                                "mouth_open": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.5,
                                        "min": 0.0
                                    }
                                },
                                "mouth_smiling": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 0.5,
                                        "min": 0.0
                                    }
                                },
                                "natural_light": {
                                    "estimate": 1,
                                    "threshold": 1
                                },
                                "radial_distortion": {
                                    "estimate": 1,
                                    "threshold": 0
                                },
                                "red_eyes": {
                                    "estimate": 1,
                                    "threshold": 0
                                },
                                "right_eye": {
                                    "estimate": 1,
                                    "threshold": [
                                        "open"
                                    ]
                                },
                                "shoulders_position": {
                                    "estimate": 1,
                                    "threshold": [
                                        "parallel"
                                    ]
                                },
                                "smile_properties": {
                                    "estimate": 1,
                                    "threshold": [
                                        "none"
                                    ]
                                },
                                "specularity_quality": {
                                    "estimate": 1,
                                    "threshold": {
                                        "max": 1.0,
                                        "min": 0.3
                                    }
                                }
                            },
                            "estimate": 0,
                            "filter": 0
                        },
                        "liveness_states": null,
                        "mask_states": null,
                        "multiface_policy": 1,
                        "pitch_threshold": null,
                        "roll_threshold": null,
                        "yaw_threshold": null
                    },
                    "extract_policy": {
                        "extract_basic_attributes": 1,
                        "fd_score_threshold": 0.0
                    },
                    "storage_policy": {
                        "attribute_policy": {
                            "store_attribute": 1
                        },
                        "face_sample_policy": {
                            "store_sample": 1
                        }
                    },
                    "verification_threshold": 0.5
                },
                "verifier_id": ...,
                "version": 0
            }
        ]

        """
        queries = {"description": description, "page": page, "page_size": pageSize, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/verifiers", "GET", queryParams=queries, **kwargs)

    def verifyRaw(
        self, candidates: List[BinaryReference], references: List[BinaryReference], verifierId: str, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Verify raw API allows to match each candidate with each reference descriptors.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/rawVerification

        Args:
            verifierId: verifier id
            candidates: list of BinaryReference objects for candidates
            references: list of BinaryReference objects for references
                        verifierId: verifier id
        Returns:
            LunaResponse: body contains json with results of matching.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>> from luna3.python_matcher.match_objects import RawDescriptorReference, RawDescriptor
        >>> from testData.resources import DESCRIPTOR
        >>>
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="test_verifier", policies=policies, raiseError=True
        ... ).json["verifier_id"]
        >>> candidate = RawDescriptorReference(referenceId="candidate_id", descriptor=DESCRIPTOR)
        >>> reference = RawDescriptorReference(referenceId="reference_id", descriptor=DESCRIPTOR)
        >>> response = client.verifyRaw(
        ...     verifierId=verifierId, candidates=[candidate], references=[reference]
        ... )
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "matches": [
                {
                    "matches": [
                        {
                            "candidate_id": "candidate_id",
                            "similarity": 1.0,
                            "status": true
                        }
                    ],
                    "reference_id": "reference_id"
                }
            ]
        }

        """
        data = {
            "candidates": [reference.asDict(encode=False) for reference in candidates],
            "references": [reference.asDict(encode=False) for reference in references],
        }
        payload = RequestPayload.buildMsgpack(body=data)
        return self.makeRequest(f"{self.baseUri}/verifiers/{verifierId}/raw", "POST", body=payload, **kwargs)

    def verifyData(
        self,
        verifierId: str,
        inputData: Union[BinaryImage, BinaryDescriptorData, List[Union[BinaryImage, BinaryDescriptorData]]],
        eventIds: Optional[List[str]] = None,
        externalIds: Optional[List[str]] = None,
        faceIds: Optional[List[str]] = None,
        attributeIds: Optional[List[str]] = None,
        imageType: Optional[int] = None,
        useExifInfo: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Perform data verification against specified objects.
        Only one objects set is allowed.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/postVerifier

        Args:
            verifierId: verifier id
            inputData: one of:
                        - Single image
                        - Single descriptor
                        - List of images and descriptors
            faceIds: faces ids to verify against
            eventIds: event ids to verify against
            externalIds: external ids to verify against
            attributeIds: temporary attribute ids to verify against
            imageType: image type (0 - raw image, 1 - face warped image)
            useExifInfo: whether to use exif info for auto orientation

        Returns:
            LunaResponse: body contains json with verifications info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import VerifierPolicies, VerifierAttributePolicy, VerifierDetectPolicy
        >>> from luna3.common.http_objs import VerifierExtractPolicy, VerifierStoragePolicy, VerifierFaceSamplePolicy
        >>> from testData.resources import FACE_WARP as image
        >>>
        >>> sampleId = client.detectFaces(
        ...     inputData=image, raiseError=True
        ... ).json["images"][0]["detections"]["samples"][0]["face"]["sample_id"]
        >>> attributeId = client.extractAttrFromSample(sampleIds=[sampleId], raiseError=True).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId, raiseError=True).json["face_id"]
        >>> policies = VerifierPolicies(
        ...     verificationThreshold=0.5,
        ...     detectPolicy=VerifierDetectPolicy(estimateEmotions=1),
        ...     extractPolicy=VerifierExtractPolicy(extractBasicAttributes=1),
        ...     storagePolicy=VerifierStoragePolicy(
        ...         faceSamplePolicy=VerifierFaceSamplePolicy(storeSample=1),
        ...         attributePolicy=VerifierAttributePolicy(storeAttribute=1),
        ...     ),
        ... )
        >>> verifierId = client.createVerifier(
        ...     description="test_verifier", policies=policies, raiseError=True).json["verifier_id"]
        >>> response = client.verifyData(
        ...     verifierId=verifierId, faceIds=[faceId], inputData=image, imageType=1, useExifInfo=1
        ... )
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "images": [
                {
                    "detections": {
                        "face_detections": [
                            {
                                "face_attributes": {
                                    "attribute_id": ...,
                                    "basic_attributes": {
                                        "age": ...,
                                        "ethnicities": {
                                            "estimations": {
                                                "african_american": ...,
                                                "asian": ...,
                                                "caucasian": ...,
                                                "indian": ...
                                            },
                                            "predominant_ethnicity": "caucasian"
                                        },
                                        "gender": 0
                                    },
                                    "samples": [...],
                                    "score": ...,
                                    "url": "/6/attributes/..."
                                },
                                "sample": {
                                    "face": {
                                        "attributes": {
                                            "emotions": {
                                                "estimations": {
                                                    "anger": ...,
                                                    "disgust": ...,
                                                    "fear": ...,
                                                    "happiness": ...,
                                                    "neutral": ...,
                                                    "sadness": ...,
                                                    "surprise": ...
                                                },
                                                "predominant_emotion": "neutral"
                                            }
                                        },
                                        "sample_id": ...,
                                        "url": "/6/samples/faces/..."
                                    }
                                },
                                "verifications": [
                                    {
                                        "face": {
                                            "face_id": ...
                                        },
                                        "similarity": ...,
                                        "status": true
                                    }
                                ]
                            }
                        ],
                        "filtered_detections": {
                            "face_detections": []
                        }
                    },
                    "error": {
                        "desc": "Success",
                        "detail": "Success",
                        "error_code": 0,
                        "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-0"
                    },
                    "filename": "raw_image",
                    "status": 1
                }
            ]
        }

        """
        payload = functions.createPayloadWithImages(inputData, {}, self.multipartBoundary)
        params = {
            "face_ids": faceIds,
            "external_ids": externalIds,
            "event_ids": eventIds,
            "attribute_ids": attributeIds,
            "image_type": imageType,
            "use_exif_info": useExifInfo,
        }

        return self.makeRequest(
            f"{self.baseUri}/verifiers/{verifierId}/verifications",
            "POST",
            body=payload,
            queryParams=params,
            **kwargs,
        )
