"""
Module realize tasks mixin
"""

from typing import Awaitable, List, Literal, Optional, Union

from luna3.common.http_objs import EstimatorHandler, EstimatorHandlerExtant
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.lunavl.mixins.base import MixinStub
from luna3.python_matcher.match_objects import EventFilters, FaceFilters
from luna3.tasks.http_objs import (
    EstimatorFTPSource,
    EstimatorNetworkDiskSource,
    EstimatorS3Source,
    EstimatorSambaSource,
    EstimatorZipSource,
    TaskNotificationPolicy,
)
from luna3.tasks.utils import getTaskPayload, getTaskSchedulePayload


class TasksMixin(MixinStub):
    """Tasks Mixin to work with tasks."""

    def getTasks(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        scheduleId: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tasks with filters.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTask

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            page: page
            pageSize: page size
            accountIdFilter: account id to get tasks
            scheduleId: schedule id filter

        Returns:
            LunaResponse: body contains json with tasks info.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415926", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415926"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(
        ...     threshold=0.5, filters=faceFilters, raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTasks(taskType=2, taskStatus=5, taskIds=[taskId])
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json["tasks"], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        [
            {
                "account_id": ...,
                "content": {
                    "descriptor": {
                        "type": "face"
                    },
                    "filters": {
                        "account_id": ...,
                        "external_ids": [
                            "14159261415926"
                        ],
                        "face_ids": [...],
                        "list_id": ...,
                        "user_data": "clustering_faces"
                    },
                    "limit": ...,
                    "objects_type": "faces",
                    "threshold": 0.5
                },
                "count_task_parts_all": 1,
                "count_task_parts_done": 1,
                "create_time": ...,
                "description": "",
                "end_time": ...,
                "last_update_time": ...,
                "notification_policy": {
                    "enable": 0,
                    "task_callbacks": []
                },
                "result_id": ...,
                "task_id": ...,
                "task_status": 5,
                "task_type": 2
            }
        ]

        """
        query = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "page": page,
            "page_size": pageSize,
            "account_id": accountIdFilter,
            "schedule_id": scheduleId,
        }
        return self.makeRequest(f"{self.baseUri}/tasks", "GET", queryParams=query, **kwargs)

    def getTaskSubTasks(
        self, taskId: int, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task's sub tasks.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getSubTasks

        Args:
            taskId: task id
            accountIdFilter: account id to get subtasks

        Returns:
            LunaResponse: body contains json with subtasks info.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415926", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415926"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(filters=faceFilters, raiseError=True).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTaskSubTasks(taskId=taskId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json["subtasks"], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        [
            {
                "content": {
                    "descriptor": {
                        "type": "face"
                    },
                    "filters": {
                        "account_id": ...,
                        "external_ids": [
                            "14159261415926"
                        ],
                        "face_ids": [...],
                        "list_id": ...,
                        "user_data": "clustering_faces"
                    },
                    "limit": ...,
                    "objects_type": "faces",
                    "threshold": ...
                },
                "create_time": ...,
                "end_time": ...,
                "result_id": ...,
                "subtask_id": ...,
                "subtask_status": 4,
                "task_id": ...
            }
        ]

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/subtasks", "GET", queryParams=queries, **kwargs)

    def getTasksCount(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        scheduleId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tasks count with filters.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTasksCount

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            accountIdFilter: account id to get tasks count
            scheduleId: schedule id filter

        Returns:
            LunaResponse: body contains json with tasks count.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415926", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415926"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(filters=faceFilters, raiseError=True).json["task_id"]
        >>> response = client.getTasksCount(taskStatus=5)
        >>> print(response.statusCode)
        200
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'count': ...}
        >>> response = client.getTasksCount(taskIds=[taskId])
        >>> print(response.json)
        {'count': 1}

        """
        query = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "account_id": accountIdFilter,
            "schedule_id": scheduleId,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/count", "GET", queryParams=query, **kwargs)

    def getTask(
        self, taskId: int, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task by id.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTask

        Args:
            taskId: task id
            accountIdFilter: account id to get task

        Returns:
            LunaResponse: body contains json with task info.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415926", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415926"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(filters=faceFilters, raiseError=True).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTask(taskId=taskId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "content": {
                "descriptor": {
                    "type": "face"
                },
                "filters": {
                    "account_id": ...,
                    "external_ids": [
                        "14159261415926"
                    ],
                    "face_ids": [...],
                    "list_id": ...,
                    "user_data": "clustering_faces"
                },
                "limit": ...,
                "objects_type": "faces",
                "threshold": ...
            },
            "count_task_parts_all": 1,
            "count_task_parts_done": 1,
            "create_time": ...,
            "description": "",
            "end_time": ...,
            "last_update_time": ...,
            "notification_policy": {
                "enable": 0,
                "task_callbacks": []
            },
            "result_id": ...,
            "task_id": ...,
            "task_status": 5,
            "task_type": 2
        }

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "GET", queryParams=queries, **kwargs)

    def getTaskErrors(
        self,
        taskId: int,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task errors by task id.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getErrors

        Args:
            taskId: task id
            page: page
            pageSize: page size
            accountIdFilter: account id to get task errors

        Returns:
            LunaResponse: body contains json with info about errors of a task by the task ID.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> unknownFaceId = "668ea250-9341-484f-a103-8dda6994a654"
        >>> faceFilters = FaceFilters(faceIds=[unknownFaceId])
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTaskErrors(taskId=taskId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "errors": [
                {
                    "additional_info": ...,
                    "description": "Object not found",
                    "detail": "Face with id '668ea250-9341-484f-a103-8dda6994a654' not found",
                    "error_code": 22002,
                    "error_id": ...,
                    "error_time": ...,
                    "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-22002",
                    "subtask_id": ...,
                    "task_id": ...
                }
            ]
        }

        """
        query = {"page": page, "page_size": pageSize, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/errors", "GET", queryParams=query, **kwargs)

    def getErrors(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get errors with filters.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTasksErrors

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter
            page: page
            pageSize: page size
            accountIdFilter: account id to get errors

        Returns:
            LunaResponse: body contains json with info about errors of a task.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> unknownFaceId = "668ea250-9341-484f-a103-8dda6994a654"
        >>> faceFilters = FaceFilters(faceIds=[unknownFaceId])
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> errorCode = client.getTaskErrors(taskId=taskId, raiseError=True).json["errors"][0]["error_code"]
        >>> response = client.getErrors(errorCode=errorCode)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json["errors"][0], sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "additional_info": ...,
            "description": "Object not found",
            "detail": "Face with id '668ea250-9341-484f-a103-8dda6994a654' not found",
            "error_code": 22002,
            "error_id": ...,
            "error_time": ...,
            "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-22002",
            "subtask_id": ...,
            "task_id": ...
        }

        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
            "page": page,
            "page_size": pageSize,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/errors", "GET", queryParams=query, **kwargs)

    def getErrorsCount(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get errors count with filters.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTasksErrorsCount  # pylint: disable=line-too-long

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter
            accountIdFilter: account id to get errors count

        Returns:
            LunaResponse: body contains json with errors count.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> unknownFaceId = "668ea250-9341-484f-a103-8dda6994a654"
        >>> faceFilters = FaceFilters(faceIds=[unknownFaceId])
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> errorCode = client.getTaskErrors(taskId=taskId, raiseError=True).json["errors"][0]["error_code"]
        >>> response = client.getErrorsCount(errorCode=errorCode)
        >>> print(response.statusCode)
        200
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'count': ...}

        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/errors/count", "GET", queryParams=query, **kwargs)

    def getError(
        self, errorId: int, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get error by id.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTaskError

        Args:
            errorId: error id
            accountIdFilter: account id to get error

        Returns:
            LunaResponse: body contains json with info about errors of a task.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> unknownFaceId = "668ea250-9341-484f-a103-8dda6994a654"
        >>> faceFilters = FaceFilters(faceIds=[unknownFaceId])
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> errorId = client.getTaskErrors(taskId=taskId, raiseError=True).json["errors"][0]["error_id"]
        >>> response = client.getError(errorId=errorId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "additional_info": ...,
            "description": "Object not found",
            "detail": "Face with id '668ea250-9341-484f-a103-8dda6994a654' not found",
            "error_code": 22002,
            "error_id": ...,
            "error_time": ...,
            "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-22002",
            "subtask_id": ...,
            "task_id": ...
        }

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/errors/{errorId}", "GET", queryParams=queries, **kwargs)

    def createLinkerTask(
        self,
        filters: Union[FaceFilters, EventFilters],
        listId: Optional[str] = None,
        createList: Optional[int] = None,
        listUserData: Optional[str] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start linker task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createLinkerTask

        Args:
            filters: required one of: face or event filters
            listId: list id to link faces to | WARNING required one of: listId or createList(1)
            createList: if 1 create list to link faces to or 0 to link them to list with id equals listId
                                                                    | WARNING required one of: listId or createList(1)
            listUserData: user data of list to be created when 'createList' is set to 1
            description: description
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(userData="link_faces", raiseError=True).json["face_id"]
        >>> faceFilters = FaceFilters(faceIds=[faceId], userData="link_faces")
        >>> response = client.createLinkerTask(
        ...     listId=listId,
        ...     filters=faceFilters,
        ...     listUserData="my_awesome_list",
        ...     description="my_LinkerTask"
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="linker",
            description=description,
            filters=filters,
            listId=listId,
            createList=createList,
            listUserData=listUserData,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/linker", "POST", body=payload, **kwargs)

    def createClusteringTask(
        self,
        filters: Union[FaceFilters, EventFilters],
        useTrackInfo: Optional[int] = None,
        descriptorTypeFilter: Optional[str] = None,
        threshold: Optional[float] = None,
        description: Optional[str] = None,
        limit: Optional[int] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start clustering task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createClusteringTask  # pylint: disable=line-too-long

        Args:
            descriptorTypeFilter: descriptor type filter, one of: face or body
            threshold: threshold
            filters: Either FaceFilters or EventFilters
            useTrackInfo: whether to use track info during the task
            description: description
            limit: candidate limit for each clutser
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> from testData.resources import FACE_WARP
        >>>
        >>> sampleId = client.saveFaceSample(FACE_WARP, raiseError=True).json["sample_id"]
        >>> attributeId = client.extractAttrFromSample(
        ...     sampleIds=[sampleId], extractBasicAttributes=1, raiseError=True
        ... ).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId, raiseError=True).json["face_id"]
        >>> faceFilters = FaceFilters(faceIds=[faceId])
        >>> response = client.createClusteringTask(filters=faceFilters, descriptorTypeFilter="face")
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="clustering",
            description=description,
            filters=filters,
            limit=limit,
            useTrackInfo=useTrackInfo,
            descriptorTypeFilter=descriptorTypeFilter,
            threshold=threshold,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/clustering", "POST", body=payload, **kwargs)

    def createGcTask(
        self,
        target: str,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        handlerId: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        storeResults: Optional[bool] = None,
        removeSamples: Optional[bool] = None,
        removeImageOrigins: Optional[bool] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start gc task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createGCTask

        Args:
            target: garbage collection target: "events" to delete events
            handlerId: handler id
            userData: user data for faces filtering
            listId: list id for faces filtering
            createTimeLt: event create time upper excluding bound
            createTimeGte: face create time bottom excluding bound
            insertTimeLt: event insert time upper excluding bound
            storeResults: whether to store task results
            removeSamples: whether to remove samples along with the event or face
            removeImageOrigins: whether to remove image origins along with the event
            description: task description
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> from luna3.common.http_objs import DetectPolicy, Policies, ExtractPolicy
        >>> from testData.resources import FACE_WARP as image
        >>>
        >>> sampleId = client.saveFaceSample(image=image, raiseError=True).json["sample_id"]
        >>> attributeId = client.extractAttrFromSample(
        ...     sampleIds=[sampleId], extractBasicAttributes=1, raiseError=True
        ... ).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId, raiseError=True).json["face_id"]
        >>> faceFilters = FaceFilters(faceIds=[faceId])
        >>> policies = Policies(detectPolicy=DetectPolicy(detectFace=1), extractPolicy=ExtractPolicy())
        >>> handlerId = client.createHandler(
        ...     description="create_Gc_task", policies=policies, raiseError=True
        ... ).json["handler_id"]
        >>> event = client.emitEvents(handlerId, inputData=image, raiseError=True)
        >>> response = client.createGcTask(
        ...     target="events",
        ...     handlerId=handlerId,
        ...     storeResults=True,
        ...     removeSamples=True,
        ...     removeImageOrigins=True,
        ...     description="my_GcTask"
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="gc",
            description=description,
            target=target,
            createTimeLt=createTimeLt,
            createTimeGte=createTimeGte,
            insertTimeLt=insertTimeLt,
            handlerId=handlerId,
            userData=userData,
            listId=listId,
            storeResults=storeResults,
            removeSamples=removeSamples,
            removeImageOrigins=removeImageOrigins,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/gc", "POST", body=payload, **kwargs)

    def createReportTask(
        self,
        taskId: int,
        columns: Optional[List[str]] = None,
        csvDelimiter: str = ",",
        saveImages: Optional[int] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start clustering task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createReporterTask

        Args:
            taskId: source task id
            columns: columns in csv report.
                    For faces ["user_data", "create_time", "lists", "external_id", "attribute_id",
                              "avatar", "event_id"]
                    For Events: ["create_time", "external_id", "handler_id", "attribute_id", "source", "stream_id,
                                "top_match", "end_time", "match_result", "extract_result", "face_id", "attach_result",
                                "gender", "age", "emotion", "mask", "ethnic_group", "tags", "user_data"]
            csvDelimiter: csv delimiter
            saveImages: whether to save images in a report
            description: description
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> faceFilters = FaceFilters(accountId=client.accountId, userData="task_clustering_faces")
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.createReportTask(
        ...     taskId=taskId,
        ...     columns=["user_data", "create_time"],
        ...     description="report_task",
        ...     csvDelimiter="$"
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        taskContent = {"content": {"filters": {"task_id": taskId}, "csv_delimiter": csvDelimiter}}
        if description is not None:
            taskContent["description"] = description
        if columns is not None:
            taskContent["content"]["columns"] = columns
        if saveImages is not None:
            taskContent["content"]["save_images"] = saveImages
        if notificationPolicy is not None:
            taskContent["notification_policy"] = notificationPolicy.asDict()

        payload = RequestPayload.buildJson(body=taskContent)
        return self.makeRequest(f"{self.baseUri}/tasks/reporter", "POST", body=payload, **kwargs)

    def createLambdaTask(
        self,
        content: dict,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start lambda task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createLambdaTask

        Args:
            content: task content
            description: task description
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.
        """
        payload = getTaskPayload(
            taskType="lambda",
            description=description,
            content=content,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/lambda", "POST", body=payload, **kwargs)

    def getTaskResult(
        self, taskId: int, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task result by task id. Optional you can set, that task must belong to account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTaskResult

        Args:
            taskId: task id
            accountIdFilter: account id to get task result

        Returns:
            LunaResponse: body contains json with task result.

        >>> import json
        >>> import time
        >>> from testData.resources import FACE_WARP
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> sampleId = client.saveFaceSample(FACE_WARP, raiseError=True).json["sample_id"]
        >>> attributeId = client.extractAttrFromSample(
        ...     sampleIds=[sampleId], extractBasicAttributes=1, raiseError=True
        ... ).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId, raiseError=True).json["face_id"]
        >>> faceFilters = FaceFilters(faceIds=[faceId])
        >>> taskId = client.createClusteringTask(
        ...     filters=faceFilters, descriptorTypeFilter="face", raiseError=True
        ... ).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTaskResult(taskId=taskId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True))  # doctest: +ELLIPSIS
        {"errors": [...], "result": {"clusters": [[...]], "total_clusters": 1, "total_objects": 1}}

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/result", "GET", queryParams=queries, **kwargs)

    def cancelTask(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Cancel a task by the task id. Optional you can set, that task must belong to account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/cancelTask

        Args:
            taskId: task id

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import EventFilters
        >>>
        >>> eventFilters = EventFilters(accountId=client.accountId)
        >>> taskId = client.createClusteringTask(filters=eventFilters, raiseError=True).json["task_id"]
        >>> response = client.cancelTask(taskId=taskId)
        >>> status = client.getTask(taskId).json["task_status"]
        >>> status in (2, 5) # task done or cancel
        True

        """
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "PATCH", **kwargs)

    def deleteTask(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete a task by the task id. Optional you can set, that task must belong to account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/deleteTask

        Args:
            taskId: task id

        Returns:
            LunaResponse: not contains response body.


        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> taskId = client.createGcTask(target="events", raiseError=True).json["task_id"]
        >>> response = client.deleteTask(taskId=taskId)
        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "DELETE", **kwargs)

    def createCrossMatchTask(
        self,
        referenceFilters: Union[FaceFilters, EventFilters],
        candidateFilters: Union[FaceFilters, EventFilters],
        descriptorTypeFilter: Optional[str] = None,
        threshold: Optional[float] = None,
        limit: Optional[int] = None,
        description: Optional[str] = None,
        accountIdFilter: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start clustering task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createCrossMatchTask  # pylint: disable=line-too-long

        Args:
            threshold: threshold
            limit: maximum candidates count per reference
            referenceFilters: filters for getting references from either faces or events
            candidateFilters: filters for getting candidates from either faces or events
            descriptorTypeFilter: descriptor type filter, one of: face or body
            description: description
            accountIdFilter: account id filter
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters, EventFilters
        >>> from luna3.common.http_objs import Policies, DetectPolicy, ExtractPolicy
        >>> from testData.resources import FACE_WARP as image
        >>>
        >>> sampleId = client.saveFaceSample(image=image, raiseError=True).json["sample_id"]
        >>> attributeId = client.extractAttrFromSample(
        ...     sampleIds=[sampleId], extractBasicAttributes=1, raiseError=True
        ... ).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId, raiseError=True).json["face_id"]
        >>> faceFilters = FaceFilters(faceIds=[faceId])
        >>> policies = Policies(detectPolicy=DetectPolicy(detectFace=1), extractPolicy=ExtractPolicy())
        >>> handlerId = client.createHandler(
        ...     description="test_handler_tasks", policies=policies, raiseError=True
        ... ).json["handler_id"]
        >>> eventId = client.emitEvents(handlerId, inputData=image, raiseError=True).json["events"][0]["event_id"]
        >>> eventFilters = EventFilters(eventIds=[eventId])
        >>> response = client.createCrossMatchTask(
        ...     referenceFilters=faceFilters,
        ...     candidateFilters=eventFilters,
        ...     descriptorTypeFilter="face",
        ...     threshold=0.5,
        ...     limit=1,
        ...     description="matching_visitors"
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="cross_match",
            referenceFilters=referenceFilters,
            candidateFilters=candidateFilters,
            descriptorTypeFilter=descriptorTypeFilter,
            accountIdFilter=accountIdFilter,
            threshold=threshold,
            limit=limit,
            description=description,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/cross_match", "POST", body=payload, **kwargs)

    def createROCTask(
        self,
        markup: List[dict],
        accountIdFilter: Optional[str] = None,
        thresholdHitTop: Optional[float] = None,
        limit: Optional[int] = None,
        keyFPRs: Optional[List[float]] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start ROC task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createROCTask

        Args:
            markup: markup list of dicts with face id and label (group id)
            accountIdFilter: account id filter
            thresholdHitTop: similarity threshold to calculate hit into the top-N
            limit: maximum candidates count per reference
            keyFPRs: roc-curve key FRP value list
            description: description
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import FACE_WARP
        >>>
        >>> sampleId = client.saveFaceSample(FACE_WARP, raiseError=True).json["sample_id"]
        >>> attributeId = client.extractAttrFromSample(
        ...     sampleIds=[sampleId], extractBasicAttributes=1, raiseError=True
        ... ).json[0]["attribute_id"]
        >>> faceId = client.createFace(attributeId=attributeId).json["face_id"]
        >>> faceId_2 = client.createFace(attributeId=attributeId).json["face_id"]
        >>> response = client.createROCTask(
        ...     markup=[{"face_id": faceId, "label": 0}, {"face_id": faceId_2, "label": 1}],
        ...     description="matching_people",
        ...     thresholdHitTop=0.95,
        ...     limit=2,
        ...     keyFPRs=[0.000001, 0.001]
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="roc",
            markup=markup,
            accountIdFilter=accountIdFilter,
            thresholdHitTop=thresholdHitTop,
            limit=limit,
            keyFPRs=keyFPRs,
            description=description,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/roc", "POST", body=payload, **kwargs)

    def createExporterTask(
        self,
        filters: Union[FaceFilters, EventFilters],
        columns: Optional[List[str]] = None,
        saveImages: Optional[int] = None,
        csvDelimiter: str = ",",
        description: Optional[str] = None,
        descriptorTypeFilter: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start exporter task.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createExporterTask

        Args:
            filters: face/event filters
            columns: columns in csv report.
                    For faces ["face_id", "user_data", "create_time", "lists", "external_id", "attribute_id",
                              "avatar", "event_id"]
                    For Events: ["event_id", "create_time", "end_time" "external_id", "handler_id", "attribute_id",
                                "source", "stream_id", "top_match", "match_result", "extract_result", "face_id",
                                "attach_result", "gender", "age", "emotion", "mask", "ethnic_group", "tags",
                                "user_data"]
            csvDelimiter: csv delimiter
            saveImages: whether to save images in a report
            description: description
            descriptorTypeFilter: descriptor type filter, one of: face or body
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> response = client.createExporterTask(
        ...     description="report_task_one",
        ...     csvDelimiter="$",
        ...     filters=FaceFilters(accountId=client.accountId),
        ...     columns=["face_id", "user_data", "create_time", "external_id", "lists", "avatar", "event_id"]
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        payload = getTaskPayload(
            taskType="exporter",
            filters=filters,
            columns=columns,
            saveImages=saveImages,
            csvDelimiter=csvDelimiter,
            description=description,
            descriptorTypeFilter=descriptorTypeFilter,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/exporter", "POST", body=payload, **kwargs)

    def createEstimatorTask(
        self,
        source: Union[
            EstimatorZipSource, EstimatorS3Source, EstimatorNetworkDiskSource, EstimatorFTPSource, EstimatorSambaSource
        ],
        handler: Union[EstimatorHandler, EstimatorHandlerExtant, str],
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start estimator task

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createEstimatorTask  # pylint: disable=line-too-long

        Args:
            source: estimator task source
            description: task description
            handler: handler for image processing, handler id or  description + policies
            notificationPolicy: task notification policy

        Returns:
            LunaResponse: body contains json with task id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.common.http_objs import Policies, DetectPolicy, ExtractPolicy, StoragePolicy, EventPolicy
        >>> from luna3.common.http_objs import NotificationPolicy, FacePolicy, LinkToListPolicy, EstimatorHandler
        >>> from luna3.tasks.http_objs import EstimatorZipSource
        >>>
        >>> listId = client.createList(raiseError=True).json["list_id"]
        >>> policies = Policies(
        ...     detectPolicy=DetectPolicy(estimateEmotions=1, estimateMask=1),
        ...     extractPolicy=ExtractPolicy(extractBasicAttributes=1, extractFaceDescriptor=1, fdScoreThreshold=0.0),
        ...     storagePolicy=StoragePolicy(
        ...         eventPolicy=EventPolicy(storeEvent=0),
        ...         notificationPolicy=NotificationPolicy(sendNotification=0),
        ...         facePolicy=FacePolicy(storeFace=1, linkToListsPolicy=[LinkToListPolicy(listId=listId)]),
        ...     ),
        ... )
        >>> response = client.createEstimatorTask(
        ...     handler=EstimatorHandler(policies, "luna3_estimator_task_example"),
        ...     source=EstimatorZipSource(url="http://127.0.0.1:5000/archive")
        ... )
        >>> print(response.statusCode)
        202
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'task_id': ...}

        """
        if isinstance(handler, str):
            handler = EstimatorHandlerExtant(handlerId=handler)
        payload = getTaskPayload(
            taskType="estimator",
            handler=handler,
            source=source,
            description=description,
            withoutAccountFilters=True,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/estimator", "POST", body=payload, **kwargs)

    def createSchedule(
        self,
        taskType: Literal[1, 2, 3, 4, 6, 7, 8, 9, 10],
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        taskNotificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create schedule

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createSchedule  # pylint: disable=line-too-long

        Args:
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task
            taskNotificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     ),
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ... )
        >>> print(response.statusCode)
        201
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'schedule_id': ...}

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            withoutAccountFilters=True,
            taskNotificationPolicy=taskNotificationPolicy,
            **contentKwargs,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/schedules", "POST", body=payload, **kwargs)

    def putSchedule(
        self,
        scheduleId: int,
        taskType: int,
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        taskNotificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update schedule

        Args:
            scheduleId: schedule id
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task
            taskNotificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     )
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ...     raiseError=True
        ... )
        >>> scheduleId = response.json["schedule_id"]
        >>> response = client.putSchedule(
        ...     scheduleId=scheduleId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="0 * * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ... )
        >>> print(response.statusCode)
        200
        >>> print(response.json)  # doctest: +ELLIPSIS
        {'schedule_id': ...}

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            withoutAccountFilters=True,
            taskNotificationPolicy=taskNotificationPolicy,
            **contentKwargs,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "PUT", body=payload, **kwargs)

    def patchSchedule(
        self,
        scheduleId: int,
        action: str,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch schedule by id

        Args:
            scheduleId: schedule id
            action: action for schedule

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     )
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ...     raiseError=True
        ... )
        >>> scheduleId = response.json["schedule_id"]
        >>> response = client.patchSchedule(
        ...     scheduleId=scheduleId,
        ...     action="pause"
        ... )
        >>> print(response.statusCode)
        204

        """
        query = {"action": action}
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "PATCH", queryParams=query, **kwargs)

    def getSchedules(
        self,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedules with filters

        Args:
            createTimeLt: schedule create time upper excluding bound
            createTimeGte: schedule create time lower including bound
            taskType: task type filter
            accountIdFilter: account id to get schedules
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     )
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ...     raiseError=True
        ... )
        >>> scheduleId = response.json["schedule_id"]
        >>> response = client.getSchedules(
        ...                 accountIdFilter=client.accountId
        ... )
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4)) # doctest: +ELLIPSIS
        [
            {
                "account_id": ...,
                "create_time": ...,
                "last_update_time": ...,
                "next_run_time": ...,
                "schedule_id": ...,
                "status": "running",
                "task": {
                    "content": {
                        "create_list": 1,
                        "filters": {
                            "account_id": ...
                        },
                        "objects_type": "faces"
                    },
                    "notification_policy": {
                        "enable": 0,
                        "task_callbacks": []
                    },
                    "task_type": 1
                },
                "task_ids": [],
                "trigger": {
                    "cron": "5 4 * * *",
                    "cron_timezone": "utc"
                }
            }
        ]

        """
        query = {
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "task_type": taskType,
            "page": page,
            "page_size": pageSize,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/", "GET", queryParams=query, **kwargs)

    def getSchedule(
        self,
        scheduleId: int,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedule by id

        Args:
            scheduleId: schedule id
            accountIdFilter: account id to get schedules

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     )
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ...     raiseError=True
        ... )
        >>> scheduleId = response.json["schedule_id"]
        >>> response = client.getSchedule(
        ...     scheduleId=scheduleId
        ... )
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4)) # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "last_update_time": ...,
            "next_run_time": ...,
            "schedule_id": ...,
            "status": "running",
            "task": {
                "content": {
                    "create_list": 1,
                    "filters": {
                        "account_id": ...
                    },
                    "objects_type": "faces"
                },
                "notification_policy": {
                    "enable": 0,
                    "task_callbacks": []
                },
                "task_type": 1
            },
            "task_ids": [],
            "trigger": {
                "cron": "5 4 * * *",
                "cron_timezone": "utc"
            }
        }

        """
        query = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "GET", queryParams=query, **kwargs)

    def deleteSchedule(
        self,
        scheduleId: int,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete schedule by id

        Args:
            scheduleId: schedule id
            accountIdFilter: account id to delete schedules

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> contentKwargs=dict(
        ...             filters=FaceFilters(accountId=client.accountId), createList=1
        ...     )
        >>> response = client.createSchedule(
        ...     accountId=client.accountId,
        ...     taskType=1,
        ...     contentKwargs=contentKwargs,
        ...     cron="5 4 * * *",
        ...     cronTz="utc",
        ...     startImmediately=False,
        ...     raiseError=True
        ... )
        >>> scheduleId = response.json["schedule_id"]
        >>> response = client.deleteSchedule(
        ...     scheduleId=scheduleId
        ... )
        >>> print(response.statusCode)
        204

        """
        query = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "DELETE", queryParams=query, **kwargs)

    def getTaskNotificationPolicy(
        self, taskId: int, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task notification policy by task id.

        Args:
            taskId: task id
            accountIdFilter: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415927", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415927"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(filters=faceFilters, raiseError=True).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.getTaskNotificationPolicy(taskId=taskId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "enable": 0,
            "task_callbacks": []
        }

        """
        query = {"account_id": accountIdFilter}
        return self.makeRequest(
            f"{self.baseUri}/tasks/{taskId}/notification_policy", "GET", queryParams=query, **kwargs
        )

    def replaceTaskNotificationPolicy(
        self,
        taskId: int,
        accountIdFilter: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Replace task notification policy by task id.

        Args:
            taskId: task id
            accountIdFilter: account id
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        >>> import json
        >>> import time
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.python_matcher.match_objects import FaceFilters
        >>> from luna3.tasks.http_objs import TaskNotificationPolicy
        >>>
        >>> listId = client.createList("happy_persons", raiseError=True).json["list_id"]
        >>> faceId = client.createFace(
        ...     externalId="14159261415928", lists=[listId], userData="clustering_faces", raiseError=True
        ... ).json["face_id"]
        >>> faceFilters = FaceFilters(
        ...     externalIds=["14159261415928"],
        ...     listId=listId,
        ...     faceIds=[faceId],
        ...     userData="clustering_faces",
        ... )
        >>> taskId = client.createClusteringTask(filters=faceFilters, raiseError=True).json["task_id"]
        >>> time.sleep(1)
        >>> response = client.replaceTaskNotificationPolicy(taskId=taskId,
        ... notificationPolicy=TaskNotificationPolicy(enable=0, taskCallbacks=[]))
        >>> print(response.statusCode)
        204

        """
        payload = RequestPayload.buildJson(body=notificationPolicy.asDict())
        payload.applyAccountIdHeader(accountIdFilter)
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/notification_policy", "PUT", body=payload, **kwargs)
