"""
Module implements client for luna-tasks
"""

from numbers import Real
from typing import Awaitable, List, Literal, Optional, Union

import aiohttp

from ..common.base_api import BaseHTTPAPI
from ..common.http_objs import EstimatorHandler, EstimatorHandlerExtant
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from ..python_matcher.match_objects import AttributeFilters, EventFilters, FaceFilters
from .http_objs import (
    AEEventFilters,
    AEFaceFilters,
    EstimatorFTPSource,
    EstimatorNetworkDiskSource,
    EstimatorS3Source,
    EstimatorSambaSource,
    EstimatorZipSource,
    TaskNotificationPolicy,
)
from .utils import getTaskPayload, getTaskSchedulePayload


class TasksApi(BaseHTTPAPI):
    """
    Class for request to luna-tasks.

    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5050",
        api: Optional[int] = 2,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def getTasks(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        accountId: Optional[str] = None,
        scheduleId: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tasks with filters.

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            accountId: account id
            scheduleId: schedule id
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "account_id": accountId,
            "schedule_id": scheduleId,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeRequest(f"{self.baseUri}/tasks", "GET", queryParams=query, **kwargs)

    def getTasksCount(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        accountId: Optional[str] = None,
        scheduleId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tasks count with filters.

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            accountId: account id
            scheduleId: schedule id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "account_id": accountId,
            "schedule_id": scheduleId,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/count", "GET", queryParams=query, **kwargs)

    def getTask(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task by id.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "GET", queryParams=query, **kwargs)

    def getTaskNotificationPolicy(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task notification policy by task id.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"account_id": accountId}
        return self.makeRequest(
            f"{self.baseUri}/tasks/{taskId}/notification_policy", "GET", queryParams=query, **kwargs
        )

    def replaceTaskNotificationPolicy(
        self,
        taskId: int,
        accountId: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Replace task notification policy by task id.

        Args:
            taskId: task id
            accountId: account id
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        payload = RequestPayload.buildJson(body=notificationPolicy.asDict())
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/notification_policy", "PUT", body=payload, **kwargs)

    def getTaskErrors(
        self,
        taskId: int,
        accountId: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task errors by task id.

        Args:
            taskId: task id
            accountId: account id
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"account_id": accountId, "page": page, "page_size": pageSize}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/errors", "GET", queryParams=query, **kwargs)

    def getTaskErrorCount(
        self,
        taskId: int,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        errorCode: Optional[int] = None,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task errors by task id.

        Args:
            taskId: task id
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            errorCode: task status filter
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "error_code": errorCode,
            "account_id": accountId,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/errors/count", "GET", queryParams=query, **kwargs)

    def getTaskSubTasks(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task subtasks by task id.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/subtasks", "GET", queryParams=query, **kwargs)

    def getErrors(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        accountId: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get errors with filters.

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter
            accountId: account id
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
            "account_id": accountId,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/errors", "GET", queryParams=query, **kwargs)

    def getErrorsCount(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get errors with filters.

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
            "account_id": accountId,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/errors/count", "GET", queryParams=query, **kwargs)

    def getError(
        self, errorId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get error by id.

        Args:
            errorId: error id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/errors/{errorId}", "GET", queryParams=query, **kwargs)

    def taskLinker(
        self,
        accountId: str,
        filters: Union[FaceFilters, EventFilters],
        listId: Optional[str] = None,
        createList: Optional[int] = None,
        listUserData: Optional[str] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start linker task

        Args:
            accountId: account id
            description: description
            listId: list id to link faces to | WARNING required one of: listId or createList(1)
            createList: if 1 create list to link faces to or 0 to link them to list with id equals listId
                                                                    | WARNING required one of: listId or createList(1)
            listUserData: user data of list to be created when 'createList' is set to 1
            filters: required one of: face or event Filters
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="linker",
            accountId=accountId,
            description=description,
            filters=filters,
            listId=listId,
            createList=createList,
            listUserData=listUserData,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/linker", "POST", body=payload, **kwargs)

    def taskClustering(
        self,
        accountId: str,
        filters: Union[FaceFilters, EventFilters],
        useTrackInfo: Optional[int] = None,
        descriptorTypeFilter: Optional[str] = None,
        threshold: Optional[float] = None,
        description: Optional[str] = None,
        limit: Optional[int] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start clustering task

        Args:
            accountId: account id
            filters: required one of: face or event Filters
            useTrackInfo: whether to use track info during the task
            descriptorTypeFilter: descriptor type filter, one of: face or body
            threshold: threshold
            description: description
            limit: candidate limit for each cluster
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="clustering",
            accountId=accountId,
            description=description,
            filters=filters,
            limit=limit,
            useTrackInfo=useTrackInfo,
            descriptorTypeFilter=descriptorTypeFilter,
            threshold=threshold,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/clustering", "POST", body=payload, **kwargs)

    def taskGc(
        self,
        accountId: str,
        target: str,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        handlerId: Optional[str] = None,
        descriptorVersion: Optional[int] = None,
        descriptorType: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        gcAccountId: Optional[str] = None,
        storeResults: Optional[bool] = None,
        removeSamples: Optional[bool] = None,
        removeImageOrigins: Optional[bool] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start gc task

        Args:
            accountId: account id
            target: garbage collection target:
                   "face_descriptors" to delete descriptors by version from luna-faces
                   "event_descriptors" to delete descriptors by version and type from luna-events
                   "faces" to delete faces
                   "events" to delete events
            descriptorVersion: descriptor version
            descriptorType: descriptor type, "face" or "body"
            handlerId: handler id
            createTimeLt: attribute or event create time upper excluding bound
            createTimeGte: face create time bottom excluding bound
            insertTimeLt: event insert time upper excluding bound
            userData: user data for faces filtering
            listId: list id for faces filtering
            gcAccountId: account id to clear attributes with | only for "events" target
            storeResults: whether to store task results
            removeSamples: whether to remove samples along with the event or face
            removeImageOrigins: whether to remove image origins along with the event
            description: task description
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="gc",
            accountId=accountId,
            description=description,
            target=target,
            createTimeLt=createTimeLt,
            createTimeGte=createTimeGte,
            insertTimeLt=insertTimeLt,
            handlerId=handlerId,
            descriptorVersion=descriptorVersion,
            descriptorType=descriptorType,
            userData=userData,
            listId=listId,
            gcAccountId=gcAccountId,
            storeResults=storeResults,
            removeSamples=removeSamples,
            removeImageOrigins=removeImageOrigins,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/gc", "POST", body=payload, **kwargs)

    def createReport(
        self,
        accountId: str,
        taskId: int,
        csvDelimiter: str = ",",
        columns: Optional[List[str]] = None,
        description: Optional[str] = None,
        saveImages: Optional[int] = None,
        filterAccountId: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start clustering task

        Args:
            accountId: account id
            description: description
            saveImages: whether to save images in a report
            csvDelimiter: csv delimiter
            taskId: source task id
            columns: columns in csv report.
                    For faces ["user_data", "create_time", "lists", "external_id", "attribute_id",
                              "avatar", "event_id", ...]
                    For Events: ["create_time", "end_time", "external_id", "handler_id", "attribute_id", "source",
                                "stream_id", "top_match", "match_result", "extract_result", "face_id", "attach_result",
                                "gender", "age", "emotion", "ethnic_group", "tags", "user_data", ...]
            filterAccountId: account id to filter images by
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        taskContent = {
            "content": {"csv_delimiter": csvDelimiter, "filters": {"task_id": taskId}},
        }
        if columns is not None:
            taskContent["content"]["columns"] = columns
        if saveImages is not None:
            taskContent["content"]["save_images"] = saveImages
        if filterAccountId is not None:
            taskContent["content"]["filters"]["account_id"] = filterAccountId
        if description is not None:
            taskContent["description"] = description
        if notificationPolicy is not None:
            taskContent["notification_policy"] = notificationPolicy.asDict()

        payload = RequestPayload.buildJson(body=taskContent)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/tasks/reporter", "POST", body=payload, **kwargs)

    def taskAdditionalExtract(
        self,
        accountId: str,
        extractionTarget: str,
        description: Optional[str] = None,
        missing: Optional[bool] = None,
        descriptorVersion: Optional[int] = None,
        filters: Union[AEFaceFilters, AEEventFilters] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start additional extract task

        Args:
            accountId: account id (task author)
            description: task description
            extractionTarget: what part of an attribute to extract
            missing: whether extract only missing attribute parts
            descriptorVersion: descriptor version to extract | use if 'extractionTarget' is "descriptor"
            filters: object filters
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="additional_extract",
            accountId=accountId,
            description=description,
            extraction_target=extractionTarget,
            missing=missing,
            descriptor_version=descriptorVersion,
            filters=filters,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/additional_extract", "POST", body=payload, **kwargs)

    def getTaskResult(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task result by task id. Optional you can set, that task must belong to account.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` with task result will be returned.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}/result", "GET", queryParams=queries, **kwargs)

    def cancelTask(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Cancel a task by the task id. Optional you can set, that task must belong to account.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` with task result will be returned.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "PATCH", queryParams=queries, **kwargs)

    def deleteTask(
        self, taskId: int, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete a task by the task id. Optional you can set, that task must belong to account.

        Args:
            taskId: task id
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` with task result will be returned.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/{taskId}", "DELETE", queryParams=queries, **kwargs)

    def taskCrossMatch(
        self,
        accountId: str,
        referenceFilters: Union[FaceFilters, EventFilters, AttributeFilters],
        candidateFilters: Union[FaceFilters, EventFilters, AttributeFilters],
        descriptorTypeFilter: Optional[str] = None,
        accountIdFilter: Optional[str] = None,
        threshold: Optional[float] = None,
        limit: Optional[int] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start cross-match task

        Args:
            accountId: account id
            description: description
            threshold: threshold
            limit: maximum candidates count per reference
            referenceFilters: filters for getting references from either faces, attributes or events
            candidateFilters: filters for getting candidates from either faces, attributes or events
            accountIdFilter: account id filter (is applied on both referenceFilters and candidateFilters processing)
            descriptorTypeFilter: descriptor type filter, one of: face or body
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="cross_match",
            accountId=accountId,
            referenceFilters=referenceFilters,
            candidateFilters=candidateFilters,
            descriptorTypeFilter=descriptorTypeFilter,
            accountIdFilter=accountIdFilter,
            threshold=threshold,
            limit=limit,
            description=description,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/cross_match", "POST", body=payload, **kwargs)

    def taskROC(
        self,
        accountId: str,
        markup: List[dict],
        accountIdFilter: Optional[str] = None,
        thresholdHitTop: Optional[float] = None,
        limit: Optional[int] = None,
        keyFPRs: Optional[List[float]] = None,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start ROC task

        Args:
            accountId: account id
            description: description
            markup: markup list of dicts with face id and label (group id)
            accountIdFilter: account id filter
            thresholdHitTop: similarity threshold to calculate hit into the top-N
            limit: maximum candidates count per reference
            keyFPRs: roc-curve key FRP value list
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="roc",
            accountId=accountId,
            markup=markup,
            accountIdFilter=accountIdFilter,
            thresholdHitTop=thresholdHitTop,
            limit=limit,
            keyFPRs=keyFPRs,
            description=description,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/roc", "POST", body=payload, **kwargs)

    def taskExporter(
        self,
        accountId: str,
        filters: Union[FaceFilters, EventFilters],
        columns: Optional[List[str]] = None,
        saveImages: Optional[int] = None,
        csvDelimiter: str = ",",
        description: Optional[str] = None,
        descriptorTypeFilter: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start exporter task.

        Args:
            accountId: account id
            description: description
            columns: columns in csv report.
                For faces ["face_id", "user_data", "create_time", "lists", "external_id", "attribute_id",
                          "avatar", "event_id"]
                For Events: ["event_id", "create_time", "external_id", "handler_id", "attribute_id", "source",
                            "stream_id", "end_time", "top_match", "match_result", "extract_result", "face_id",
                            "attach_result", "gender", "age", "emotion", "ethnic_group", "tags", "user_data", "mask",
                            "face_detection", "body_detection"]
            filters: required one of: face or event Filters
            descriptorTypeFilter: descriptor type filter, one of: face or body
            saveImages: whether to save images in a report
            csvDelimiter: csv delimiter
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="exporter",
            accountId=accountId,
            filters=filters,
            columns=columns,
            saveImages=saveImages,
            csvDelimiter=csvDelimiter,
            description=description,
            descriptorTypeFilter=descriptorTypeFilter,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/exporter", "POST", body=payload, **kwargs)

    def taskEstimator(
        self,
        accountId: str,
        accountIdFilter: str,
        handler: Union[EstimatorHandler, EstimatorHandlerExtant],
        source: Union[
            EstimatorZipSource, EstimatorS3Source, EstimatorNetworkDiskSource, EstimatorFTPSource, EstimatorSambaSource
        ],
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start estimator task

        Args:
            accountId: account id (task author)
            accountIdFilter: account id filter
            description: task description
            handler: handler with estimation policies
            source: estimator task source
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="estimator",
            accountId=accountId,
            handler=handler,
            source=source,
            account_id_filter=accountIdFilter,
            description=description,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/estimator", "POST", body=payload, **kwargs)

    def taskLambda(
        self,
        accountId: str,
        content: dict,
        description: Optional[str] = None,
        notificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start lambda task

        Args:
            accountId: account id
            content: task content with lambda_id
            description: task description
            notificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.

        """
        payload = getTaskPayload(
            taskType="lambda",
            accountId=accountId,
            description=description,
            content=content,
            notificationPolicy=notificationPolicy,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/lambda", "POST", body=payload, **kwargs)

    def createSchedule(
        self,
        accountId: str,
        taskType: Literal[1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        taskNotificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create schedule

        Args:
            accountId: account id (task author)
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task
            taskNotificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            accountId=accountId,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            taskNotificationPolicy=taskNotificationPolicy,
            **contentKwargs,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/schedules", "POST", body=payload, **kwargs)

    def putSchedule(
        self,
        accountId: str,
        scheduleId: int,
        taskType: int,
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        taskNotificationPolicy: Optional[TaskNotificationPolicy] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update schedule

        Args:
            accountId: account id (task author)
            scheduleId: schedule id
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task
            taskNotificationPolicy: task notification policy

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            accountId=accountId,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            taskNotificationPolicy=taskNotificationPolicy,
            **contentKwargs,
        )
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "PUT", body=payload, **kwargs)

    def patchSchedule(
        self,
        accountId: str,
        scheduleId: int,
        action: str,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch schedule by id

        Args:
            accountId: account id (task author)
            scheduleId: schedule id
            action: action for schedule

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        headers = {"Luna-Account-Id": accountId}
        query = {"action": action}
        return self.makeRequest(
            f"{self.baseUri}/tasks/schedules/{scheduleId}", "PATCH", queryParams=query, headers=headers, **kwargs
        )

    def getSchedules(
        self,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        accountId: Optional[str] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedules with filters

        Args:
            createTimeLt: schedule create time upper excluding bound
            createTimeGte: schedule create time lower including bound
            taskType: task type filter
            accountId: account id
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "task_type": taskType,
            "account_id": accountId,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/", "GET", queryParams=query, **kwargs)

    def getSchedule(
        self,
        scheduleId: int,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedule by id

        Args:
            accountId: account id (task author)
            scheduleId: schedule id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {}
        if accountId is not None:
            query = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "GET", queryParams=query, **kwargs)

    def deleteSchedule(
        self,
        scheduleId: int,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete schedule by id

        Args:
            accountId: account id (task author)
            scheduleId: schedule id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {}
        if accountId is not None:
            query = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/tasks/schedules/{scheduleId}", "DELETE", queryParams=query, **kwargs)
